from pysam.libchtslib cimport *
from pysam.libcalignmentfile cimport *

import collections, array, struct, sys
import CGAT.Experiment as E


def filter(AlignmentFile genome_samfile,
           AlignmentFile output_samfile,
           AlignmentFile output_mismapped,
           AlignmentFile transcripts_samfile,
           AlignmentFile junctions_samfile,
           transcripts,
           regions = None,
           unique = False,
           remove_contigs = None,
           colour_mismatches = False,
           ignore_mismatches = False,
           ignore_junctions = True,
           ignore_transcripts = False ):
    '''
    To conserve memory, the tid and NM flag from *transcripts_samfile*
    are packed into memory. As a consequence, this method requires 
    max(NM) < 256 (2^8) and max(tid) < 16777216 (2^24)

    If *unique* is set, only uniquely matching reads will be output.

    If *remove_contigs* is given, contigs that are in remove_contigs will
    be removed. Note that this will only remove the alignment, but not
    all alignments of a particuluar read and the NH flag will *NOT* be
    updated.

    If *colour_mismatches* is set, the ``CM`` tag will be used
    to count differences. By default, the ``NM`` tag is used.
    The tag that is used needs to present in both *transcripts_samfile*
    and *genome_samfile*.

    If *ignore_mismatches* is set, the number of mismatches is ignored.

    If *regions* is given, alignments overlapping regions will be removed.

    '''
    cdef int ninput = 0
    cdef int nunmapped_genome = 0
    cdef int nunmapped_transcript = 0
    cdef int nmismapped = 0
    cdef int noutput = 0
    cdef int nnotfound = 0
    cdef int nlocation_ok = 0
    cdef int nnonunique = 0
    cdef int ntested = 0
    cdef int nremoved_contigs = 0
    cdef int nremoved_junctions = 0
    cdef int nskipped_junctions = 0
    cdef int nadded_junctions = 0
    cdef int nremoved_regions = 0

    cdef bint c_unique = unique
    cdef bint c_test_mismatches = not ignore_mismatches
    cdef bint c_test_junctions = not ignore_junctions
    cdef bint c_test_transcripts = not ignore_transcripts
    cdef bint c_test_regions = regions
    cdef int * remove_contig_tids 
    cdef int nremove_contig_tids = 0
    cdef AlignedSegment read
    cdef AlignedSegment match
    
    # build index
    # this method will start indexing from the current file position
    # if you decide
    cdef int ret = 1
    cdef int x
    cdef bam1_t * b = <bam1_t*>calloc(1, sizeof( bam1_t))
    cdef uint64_t pos
    cdef uint8_t * v
    cdef int32_t nm
    cdef int32_t nh
    cdef int tid
    cdef int transript_tid
    cdef long val
    cdef bint skip = 0
    cdef int * map_tid2tid

    cdef int32_t read_mismatches = 0

    # decide which tag to use
    cdef char * nm_tag = 'NM' 
    cdef char * cm_tag = 'CM'
    cdef char * tag

    if colour_mismatches:
        tag = cm_tag
    else:
        tag = nm_tag

    # set with junctions that are ignored
    skip_junctions = set()

    # build list of junctions
    if c_test_junctions:
        E.info("building junction read index")

        def _gen2(): return array.array('B') 
        junctions_index = collections.defaultdict(_gen2)
        ret = 1
        while ret > 0:
            ret = bam_read1(hts_get_bgzfp(junctions_samfile.htsfile),
                            b)
            if ret > 0:
                # ignore unmapped reads
                if b.core.flag & 4: continue
                qname = pysam_bam_get_qname(b)
                v = bam_aux_get(b, tag)
                nm = <int32_t>bam_aux2i(v)
                junctions_index[qname].append(nm)

        E.info( "built index for %i junction reads" % len(junctions_index))

        map_tid2tid = <int*>calloc(len(junctions_samfile.references), sizeof(int))
        
        for x, contig_name in enumerate(junctions_samfile.references):
            map_tid2tid[x] = genome_samfile.gettid(contig_name)

    if c_test_transcripts:
        E.info( "building transcriptome read index" )
        # L = 4 bytes
        def _gen(): return array.array('L') 
        transcriptome_index = collections.defaultdict(_gen)
        ret = 1
        while ret > 0:
            ret = bam_read1(hts_get_bgzfp(transcripts_samfile.htsfile),
                            b)
            if ret > 0:
                # ignore unmapped reads
                if b.core.flag & 4: continue
                qname = pysam_bam_get_qname(b)
                tid = b.core.tid
                v = bam_aux_get(b, tag)
                nm = <int32_t>bam_aux2i(v)
                transcriptome_index[qname].append( (tid << 8) | nm )

        E.info( "built index for %i transcriptome reads" % len(transcriptome_index))

    bam_destroy1( b )

    # setup list of contigs to remove:
    if remove_contigs:
        nremove_contig_tids = len(remove_contigs)
        remove_contig_tids = <int*>malloc( sizeof(int) * nremove_contig_tids )
        for x, rname in enumerate( remove_contigs):
            remove_contig_tids[x] = genome_samfile.gettid( rname )
    
    E.info( "starting filtering" )

    for read in genome_samfile:

        ninput += 1
        # if ninput > 10000: break

        # is unmapped?
        if read._delegate.core.flag & 4:
            nunmapped_genome += 1
            noutput += 1
            output_samfile.write( read )
            continue

        # optionally remove non-unique reads
        if c_unique:
            v = bam_aux_get(read._delegate, 'NH')
            if v != NULL:
                nh = <int32_t>bam_aux2i(v)
                if nh > 1: 
                    nnonunique += 1
                    continue

        # optionally remove reads matched to certain contigs
        if nremove_contig_tids:
            skip = 0
            for x from 0 <= x < nremove_contig_tids:
                if remove_contig_tids[x] == read.tid:
                    nremoved_contigs += 1
                    skip = 1
                    break
            if skip: continue

        g_contig = genome_samfile.getrname( read.tid )
                
        # optionally remove reads mapped to certain regions
        if c_test_regions:
            intervals = regions.get( g_contig, read.pos, read.aend )
            skip = 0
            for start, end, value in intervals:
                if read.overlap( start, end ):
                    nremoved_regions += 1
                    skip = 1
                    break

            if skip: continue

        if c_test_junctions:
            if read.qname in junctions_index:
                # can compress index before, depends on
                # how many reads you expect to test
                nm = min(junctions_index[read.qname])
                read_mismatches = read.opt(tag)
                if nm > read_mismatches: 
                    nremoved_junctions += 1
                    continue
                else:
                    skip_junctions.add(read.qname)
                                        
        # set mapped = True, if read is mapped to transcripts
        # set location_ok = True, if read matches in expected location
        # according to transcripts
        location_ok = False
        mapped = False

        if c_test_transcripts:
            # get transcripts that read matches to
            try:
                matches = transcriptome_index[read.qname]
            except KeyError:
                nnotfound += 1
                matches = None
                
            if matches:

                if c_test_mismatches:
                    read_mismatches = read.opt(tag)

                for val in matches:
                    transcript_tid = val >> 8
                    # ignore reads that are mapped to transcripts with
                    # more mismatches than the genomic location
                    if c_test_mismatches:
                        nm = val & 255
                        if nm > read_mismatches: continue

                    mapped = True

                    # find transcript
                    transcript_id = transcripts_samfile._getrname( transcript_tid )
                    gtfs = transcripts[transcript_id]
                    t_contig, t_start, t_end = gtfs[0].contig, gtfs[0].start, gtfs[-1].end

                    # does read map to genomic transcript location?
                    if g_contig == t_contig and t_start <= read.pos <= t_end:
                        location_ok = True
                        break
                    
        if location_ok:
            ntested += 1
            nlocation_ok += 1
            noutput += 1
            output_samfile.write( read )

        elif mapped:
            ntested += 1
            nmismapped += 1
            if output_mismapped:
                output_mismapped.write( read )

        else:
            nunmapped_transcript += 1
            noutput += 1
            output_samfile.write( read )

    # add junctions
    if c_test_junctions:
        E.info("adding junctions")
        junctions_samfile.reset()

        # rebuild contig map for junctions:
        if remove_contigs:
            remove_contig_tids = <int*>malloc( sizeof(int) * nremove_contig_tids )
            for x, rname in enumerate( remove_contigs):
                remove_contig_tids[x] = junctions_samfile.gettid( rname )
        
        for read in junctions_samfile:

            # optionally remove reads matched to certain contigs
            if nremove_contig_tids:
                skip = 0
                for x from 0 <= x < nremove_contig_tids:
                    if remove_contig_tids[x] == read.tid:
                        nremoved_contigs += 1
                        skip = 1
                    break
                if skip: 
                    nskipped_junctions += 1
                    continue

            if read.qname in skip_junctions:
                nskipped_junctions += 1
            else:
                nadded_junctions += 1
                noutput += 1
                # map tid from junction database to genome database
                read.tid = map_tid2tid[read.tid]
                output_samfile.write(read)
        
        free( map_tid2tid )

    c = E.Counter()
    c.input = ninput
    c.removed_nonunique = nnonunique
    c.removed_mismapped = nmismapped
    c.removed_contigs = nremoved_contigs
    c.removed_junctions = nremoved_junctions
    c.removed_regions = nremoved_regions
    c.skipped_junction_reads = len(skip_junctions)
    c.skipped_junctions = nskipped_junctions
    c.added_junctions = nadded_junctions
    c.output = noutput
    c.unmapped_genome = nunmapped_genome
    c.unmapped_transcript = nunmapped_transcript
    c.notfound = nnotfound
    c.location_ok = nlocation_ok 
    c.location_tested = ntested

    if nremove_contig_tids:
        free( remove_contig_tids )

    E.info( "filtering finished" )

    return c



'''
MatrixTools.py - 
======================================================

:Tags: Python

Code
----

'''
import string
import collections
import numpy
import pandas
from CGAT import IOTools as IOTools


class Matrix:

    def __init__(self):
        pass


def addOptions(parser):
    """add matrices to option parser."""

    parser.add_option("-f", "--format", dest="format", type="string",
                      help="format.")

    parser.add_option("--row-names", dest="row_names", type="string",
                      help="list of row names separated by ','.")

    parser.add_option("--col-names", dest="col_names", type="string",
                      help="list of col names separated by ','.")

    parser.add_option("--file-row-names", dest="file_row_names", type="string",
                      help="filename with row names.")

    parser.add_option("--file-col-names", dest="file_col_names", type="string",
                      help="filename with col names.")

    parser.add_option("--numeric", dest="numeric", action="store_true",
                      help="row and column titles are numeric.")

    parser.add_option("--is-asymmetric", dest="asymmetric", action="store_true",
                      help="matrix is asymmetric.")

    parser.add_option("--default-value", dest="default", type="string",
                      help="default value for missing values.")

    parser.add_option("--default-diagonal", dest="default_diagonal", type="string",
                      help="default value for missing values on diagonal.")

    parser.add_option("--input-format", dest="input_format", type="choice",
                      choices=(
                          "row-col-weight", "", "row-col-weight-replicates"),
                      help="input format.")

    parser.set_defaults(
        default="0",
        default_diagonal="0",
        format="string",
        asymmetric=False,
        is_numeric=False,
        row_names=None,
        col_names=None,
        file_row_names=None,
        file_col_names=None,
        input_format="row-col-weight",
    )


def getMapTokens(options):

    map_token2row, map_token2col = {}, {}

    if options.file_row_names:
        row_tokens = [string.split(
            x[:-1], "\t")[0] for x in open(options.file_row_names, "r").readlines()]
        for row_token in row_tokens:
            map_token2row[row_token] = len(map_token2row)

    if options.row_names:
        for x in options.row_names.split(","):
            map_token2row[x] = len(map_token2row)

    if options.file_col_names:
        col_tokens = [string.split(
            x[:-1], "\t")[0] for x in open(options.file_col_names, "r").readlines()]
        for col_token in col_tokens:
            map_token2col[col_token] = len(map_token2col)

    if options.col_names:
        for x in options.col_names.split(","):
            map_token2col[x] = len(map_token2col)

    if not options.asymmetric:
        if map_token2row and not map_token2col:
            map_token2col = map_token2row
        elif map_token2col and not map_token2row:
            map_token2row = map_token2col

    return map_token2row, map_token2col


def getMatrixFromEdges(lines, options,
                       in_map_token2row={},
                       in_map_token2col={}):
    """read matrix from lines
    """

    # remove comments
    lines = [x for x in lines if x[0] != "#" and len(x[:-1]) > 0]

    if in_map_token2row:
        map_token2row = in_map_token2row
    else:
        map_token2row = {}

    if in_map_token2col:
        map_token2col = in_map_token2col
    else:
        map_token2col = {}

    if options.format == "string":

        has_row_names = len(map_token2row) > 0
        has_col_names = len(map_token2col) > 0

        # if either row/column names are not given:
        if not map_token2row or not map_token2col:

            row_tokens = [string.split(x[:-1], "\t")[0] for x in lines]
            col_tokens = [string.split(x[:-1], "\t")[1] for x in lines]

            if options.is_numeric:
                row_tokens = list(map(float, row_tokens))
                col_tokens = list(map(float, col_tokens))
                row_tokens.sort()
                col_tokens.sort()
                row_tokens = list(map(str, row_tokens))
                col_tokens = list(map(str, col_tokens))
            else:
                row_tokens.sort()
                col_tokens.sort()

            if not has_row_names:
                for row_token in row_tokens:
                    if row_token not in map_token2row:
                        map_token2row[row_token] = len(map_token2row)
            if not has_col_names:
                for col_token in col_tokens:
                    if col_token not in map_token2col:
                        map_token2col[col_token] = len(map_token2col)

        if not options.asymmetric:
            for col_token in list(map_token2col.keys()):
                if col_token not in map_token2row:
                    map_token2row[col_token] = len(map_token2row)
            map_token2col = map_token2row

        matrix = [[options.default for j in range(len(map_token2col))] for i in range(
            len(map_token2row))]

        if len(map_token2col) == len(map_token2row):
            for j in range(len(map_token2col)):
                matrix[j][j] = options.default_diagonal

        # return matrix
        m = Matrix()

        if options.input_format == "row-col-weight":
            for line in lines:
                row_token, col_token, weight = string.split(
                    line[:-1], "\t")[:3]
                matrix[map_token2row[row_token]][
                    map_token2col[col_token]] = weight
                if not options.asymmetric:
                    matrix[map_token2col[col_token]][
                        map_token2row[row_token]] = weight

        elif options.input_format == "row-col-weight-replicates":
            replicates = [[0 for j in range(len(map_token2col))]
                          for i in range(len(map_token2row))]
            for line in lines:
                row_token, col_token, weight, n = string.split(
                    line[:-1], "\t")[:4]
                matrix[map_token2row[row_token]][
                    map_token2col[col_token]] = weight
                replicates[map_token2row[row_token]][
                    map_token2col[col_token]] = int(n)
                if not options.asymmetric:
                    matrix[map_token2col[col_token]][
                        map_token2row[row_token]] = weight
                    replicates[map_token2col[col_token]][
                        map_token2row[row_token]] = int(n)
            m.mReplicates = replicates

        col_tokens = list(map_token2col.items())
        col_tokens.sort(lambda x, y: cmp(x[1], y[1]))
        row_tokens = list(map_token2row.items())
        row_tokens.sort(lambda x, y: cmp(x[1], y[1]))

        m.mMatrix = matrix
        m.mMapRow2Token = row_tokens
        m.mMapCol2Token = col_tokens
        m.mMapToken2Row = map_token2row
        m.mMapToken2Col = map_token2col

    return matrix


def buildMatrixFromLists(lists, dtype=numpy.float, default=None):
    '''build a matrix from a list of lists.

    Each list is a list of tuples (row, value).
    The columns are given by order of the lists.

    Returns matrix, row_headers
    '''

    all_rows = collections.defaultdict()
    for l in lists:
        all_rows.update([(x[0], 0) for x in l])
    for x, v in enumerate(all_rows.items()):
        all_rows[v[0]] = x

    matrix = numpy.zeros((len(all_rows), len(lists)), dtype=dtype)
    if default:
        matrix.fill(default)

    for col, l in enumerate(lists):
        for row, value in l:
            matrix[all_rows[row], col] = value
    return matrix, list(all_rows.keys())


def buildMatrixFromTables(infiles, column, column_header=0, dtype=numpy.float,
                          default=None):
    '''build a matrix from a column called *column* in a series of input files.

    If column_value is None, the first column is taken as the name of the row.

    The columns are given by order of the input files.

    returns matrix, row_headers
    '''

    lists = []
    for infile in infiles:
        data = pandas.read_table(IOTools.openFile(infile))
        lists.append(list(zip(list(data[column_header]), list(data[column]))))

    return buildMatrixFromLists(lists, dtype=dtype, default=default)


def buildMatrixFromEdges(edges,
                         in_map_token2row={},
                         in_map_token2col={},
                         is_symmetric=False,
                         missing_value=0,
                         diagonal_value=0,
                         dtype=numpy.int):
    """build a matrix from an edge-list representation.

    For example, the following list of tuples::

       [('A', 'B', 1),
        ('A', 'C', 2),
        ('B', 'C', 3)]

    will be converted to the following matrix::

         A B C
       A   1 2
       B     3
       C 

    If *is_symmetric* is set to True, the matrix is assumed to be
    symmetric and missing values will automatically be filled::

         A B C
       A   1 2
       B 1   3
       C 2 3

    If edge list may contain four elements, in which case the
    fourth element is expected to be the value of the lower
    diagonal in a symmetric matrix::

       [('A', 'B', 1, 4),
        ('A', 'C', 2, 5),
        ('B', 'C', 3, 6)]

    will yield::

         A B C
       A   1 2
       B 4   3
       C 5 6


    returns a numpy matrix and lists of row and column names.
    """

    if in_map_token2row:
        map_token2row = in_map_token2row
    else:
        map_token2row = {}

    if in_map_token2col:
        map_token2col = in_map_token2col
    else:
        map_token2col = {}

    has_row_names = len(map_token2row) > 0
    has_col_names = len(map_token2col) > 0

    # if either row/column names are not given:
    if not map_token2row or not map_token2col:

        row_tokens = sorted(list(set([x[0] for x in edges])))
        col_tokens = sorted(list(set([x[1] for x in edges])))

        if not has_row_names:
            for row_token in row_tokens:
                if row_token not in map_token2row:
                    map_token2row[row_token] = len(map_token2row)
        if not has_col_names:
            for col_token in col_tokens:
                if col_token not in map_token2col:
                    map_token2col[col_token] = len(map_token2col)

        if is_symmetric:
            for col_token in list(map_token2col.keys()):
                if col_token not in map_token2row:
                    map_token2row[col_token] = len(map_token2row)
            map_token2col = map_token2row

        matrix = numpy.matrix([missing_value] *
                              len(map_token2row) * len(map_token2col),
                              dtype=numpy.dtype)
        matrix.shape = (len(map_token2row), len(map_token2col))

        if len(map_token2col) == len(map_token2row):
            for j in range(len(map_token2col)):
                matrix[j][j] = diagonal_value

        if len(edges[0] == 3):
            if is_symmetric:
                for row, col, value in edges:
                    matrix[map_token2row[row], map_token2col[col]] = \
                        matrix[map_token2row[col], map_token2col[row]] = value
            else:
                for row, col, value in edges:
                    matrix[map_token2row[row], map_token2col[col]] = value
        elif len(edges[0] == 4):
            for row, col, value1, value2 in edges:
                matrix[map_token2row[row], map_token2col[col]] = value1
                matrix[map_token2row[col], map_token2col[row]] = value2
        else:
            raise ValueError(
                "unexpected number of elements in list, expected 3 or 4, "
                "got %i" % (len(edges[0])))

    col_tokens = list(map_token2col.items())
    col_tokens.sort(lambda x, y: cmp(x[1], y[1]))
    row_tokens = list(map_token2row.items())
    row_tokens.sort(lambda x, y: cmp(x[1], y[1]))

    return matrix, row_tokens, col_tokens



'''
IGV.py - Simple wrapper to the IGV socket interface
====================================================

:Tags: Python

This code was written by Brent Pedersen.

Downloaded from https://github.com/brentp/bio-playground/blob/master/igv/igv.py
on Nov.30 2011.

'''

import socket
import os.path as op
import os
import subprocess
import time


def startIGV(command="igv.sh", port=None):
    """start IGV on a specific port."""
    args = [command]
    if port is not None:
        args.extend(['-p', str(port)])
    process = subprocess.Popen(args,
                               stdout=subprocess.PIPE)
    time.sleep(10)
    return process


class IGV(object):
    r"""
    Simple wrapper to the IGV (http://www.broadinstitute.org/software/igv/home)
    socket interface (http://www.broadinstitute.org/software/igv/PortCommands)

    requires:

        1) you have IGV running on your machine (launch with webstart here:
                http://www.broadinstitute.org/software/igv/download)

        2) you have enabled port communication in
                View -> Preferences... -> Advanced

    Successful commands return 'OK'

    example usage:

        >>> igv = IGV()
        >>> igv.genome('hg19')
        'OK'

        >>> igv.load('http://www.broadinstitute.org/igvdata/1KG/pilot2Bams/NA12878.SLX.bam')
        'OK'
        >>> igv.go('chr1:45,600-45,800')
        'OK'

    #save as svg, png, or jpg
        >>> igv.save('/tmp/r/region.svg')
        'OK'
        >>> igv.save('/tmp/r/region.png')
        'OK'

    # go to a gene name.
        >>> igv.go('muc5b')
        'OK'
        >>> igv.sort()
        'OK'
        >>> igv.save('muc5b.png')
        'OK'

    # get a list of commands that will work as an IGV batch script.
        >>> print "\n".join(igv.commands)
        snapshotDirectory /tmp/igv
        genome hg19
        goto chr1:45,600-45,800
        snapshotDirectory /tmp/r
        snapshot region.svg
        snapshot region.png
        goto muc5b
        sort base
        snapshot muc5b.png

    Note, there will be some delay as the browser has to load the annotations
    at each step.

    """
    _socket = None
    _path = None

    def __init__(self, host='127.0.0.1', port=60151, snapshot_dir='/tmp/igv'):
        self.host = host
        self.port = port
        self.commands = []
        self.connect()
        self.set_path(snapshot_dir)

    @classmethod
    def start(cls, jnlp="igv.jnlp",
              url="http://www.broadinstitute.org/igv/projects/current/"):
        import subprocess
        p = subprocess.Popen("/usr/bin/javaws -Xnosplash %s%s" % (url, jnlp),
                             shell=True, stdout=subprocess.PIPE)
        p.wait()
        return p.returncode

    def connect(self):
        if self._socket:
            self._socket.close()
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._socket.connect((self.host, self.port))

    def go(self, position):
        return self.send('goto ' + position)
    goto = go

    def genome(self, name):
        return self.send('genome ' + name)

    def load(self, url):
        return self.send('load ' + url)

    def sort(self, option='base'):
        """
        options is one of: base, position, strand, quality, sample, and
        readGroup.
        """
        assert option in ("base", "position", "strand", "quality", "sample",
                          "readGroup")
        return self.send('sort ' + option)

    def set_path(self, snapshot_dir):
        if snapshot_dir == self._path:
            return
        if not op.exists(snapshot_dir):
            os.makedirs(snapshot_dir)

        self.send('snapshotDirectory %s' % snapshot_dir)
        self._path = snapshot_dir

    def expand(self, track):
        self.send('expand %s' % track)

    def collapse(self, track):
        self.send('collapse %s' % track)

    def clear(self):
        self.send('clear')

    def send(self, cmd):
        self.commands.append(cmd)
        self._socket.send(cmd + '\n')
        return self._socket.recv(10).rstrip('\n')

    def save(self, path=None):
        if path is not None:
            # igv assumes the path is just a single filename, but
            # we can set the snapshot dir. then just use the filename.
            dirname = op.dirname(path)
            if dirname:
                self.set_path(dirname)
            return self.send('snapshot ' + op.basename(path))
        else:
            return self.send('snapshot')
    snapshot = save

if __name__ == "__main__":
    import doctest
    doctest.testmod()



'''
fastqs2fasta.py - interleave two fastq files
=============================================

:Tags: Genomics NGS FASTQ FASTA Conversion

Purpose
-------

This script is used to interleave two :term:`fastq`-formatted files
(paired data) into a single :term:`fasta`-formatted file. Read1 is
followed by read2 in the resultant file.

:term:`fastq` files MUST be sorted by read identifier.

Usage
-----

For example::

   cgat fastqs2fasta \
         --first-fastq-file=in.fastq.1.gz \
         --second-fastq-file=in.fastq.2.gz > out.fasta

If :file:`in.fastq.1.gz` looks like this::

    @r1_from_gi|387760314|ref|NC_017594.1|_Streptococcus_saliva_#0/1
    TTCTTGTTGAATCATTTCAATTGTCTCCTTTTAGTTTTATTAGATAATAACAGCTTCTTCCACAACTTCT
    +
    ??A???ABBDDBDDEDGGFGAFHHCHHIIIDIHGIFIH=HFICIHDHIHIFIFIIIIIIHFHIFHIHHHH
    @r3_from_gi|315441696|ref|NC_014814.1|_Mycobacterium_gilvum_#0/1
    ATGAACGCGGCCGAGCAACACCGCCACCACGTGAATCGGTGGTTCTACGACTGCCCGTCGGCCTTCCACC
    +

and :file:`in.fastq.2.gz` looks like this::

    A??A?B??BDBDDDBDGGFA>CFCFIIIIIIF;HFIGHCIGHIHHEHHHIIHHFDHH-HD-IDHHHGIHG
    @r1_from_gi|387760314|ref|NC_017594.1|_Streptococcus_saliva_#0/2
    ACCTTCGTTTCCAAGGTGCAGCAGGTCAACTTGATCAAACTGCCCCTTTGAACGAAGTGAAAAAACAAAT
    +
    A????@BBDBDDADABGFGFFEHHHIEHHII@IIHIHHIDHCCIHIIIHHIEI5HIHFHIEHIH=CHHC)
    @r3_from_gi|315441696|ref|NC_014814.1|_Mycobacterium_gilvum_#0/2
    GGGAGCCTGCAGCGCCGCCGCGACTGCATCGCCGCGGCCGGCATCGTGGGATGGACGGTGCGTCAGACGC
    +
    ???A?9BBDDD5@DDDGFFGFFHIIIHHIHBFHIIHIIHHH>HEIHHFI>FFHGIIHHHDHCCFIHFIHD

then the output will be::

  >r1_from_gi|387760314|ref|NC_017594.1|_Streptococcus_saliva_#0/1
  TTCTTGTTGAATCATTTCAATTGTCTCCTTTTAGTTTTATTAGATAATAACAGCTTCTTCCACAACTTCT
  >r1_from_gi|387760314|ref|NC_017594.1|_Streptococcus_saliva_#0/2
  ACCTTCGTTTCCAAGGTGCAGCAGGTCAACTTGATCAAACTGCCCCTTTGAACGAAGTGAAAAAACAAAT
  >r3_from_gi|315441696|ref|NC_014814.1|_Mycobacterium_gilvum_#0/1
  ATGAACGCGGCCGAGCAACACCGCCACCACGTGAATCGGTGGTTCTACGACTGCCCGTCGGCCTTCCACC
  >r3_from_gi|315441696|ref|NC_014814.1|_Mycobacterium_gilvum_#0/2
  GGGAGCCTGCAGCGCCGCCGCGACTGCATCGCCGCGGCCGGCATCGTGGGATGGACGGTGCGTCAGACGC
  >r4_from_gi|53711291|ref|NC_006347.1|_Bacteroides_fragilis_#0/1
  GAGGGATCAGCCTGTTATCCCCGGAGTACCTTTTATCCTTTGAGCGATGTCCCTTCCATACGGAAACACC
  >r4_from_gi|53711291|ref|NC_006347.1|_Bacteroides_fragilis_#0/2
  CAACCGTGAGCTCAGTGAAATTGTAGTATCGGTGAAGATGCCGATTACCCGCGATGGGACGAAAAGACCC
  >r5_from_gi|325297172|ref|NC_015164.1|_Bacteroides_salanitr_#0/1
  TGCGGCGAAATACCAGCCCATGCCCCGTCCCCAGAATTCCTTGGAGCAGCCTTTGTGAGGTTCGGCTTTG
  >r5_from_gi|325297172|ref|NC_015164.1|_Bacteroides_salanitr_#0/2
  AACGGCACGCACAATGCCGACCGCTACAAAAAGGCTGCCGACTGGCTCCGCAATTACCTGGTGAACGACT


Type::

   cgat fastqs2fasta --help

for command line help.


Command line options
--------------------

'''

import sys
from future.moves.itertools import zip_longest

import CGAT.IOTools as IOTools
import CGAT.Fastq as Fastq
import CGAT.Experiment as E


class PairedReadError(Exception):

    '''
    exception raised when reads aren't paired -
    could be not sorted or files of different lengths
    '''


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv:
        argv = sys.argv

    # setup command line parser
    parser = E.OptionParser(
        version="%prog version: $Id$",
        usage=globals()["__doc__"])

    parser.add_option(
        "-a", "--first-fastq-file", dest="fastq1", type="string",
        help="supply read1 fastq file")
    parser.add_option(
        "-b", "--second-fastq-file", dest="fastq2", type="string",
        help="supply read2 fastq file")

    # add common options (-h/--help, ...) and parse command line
    (options, args) = E.Start(parser, argv=argv)

    if args and len(args) == 2:
        options.fastq1, options.fastq2 = args

    fastq1 = IOTools.openFile(options.fastq1)
    fastq2 = IOTools.openFile(options.fastq2)

    E.info("iterating over fastq files")
    f1_count = 0
    for f1, f2 in zip_longest(Fastq.iterate(fastq1),
                              Fastq.iterate(fastq2)):
        if not (f1 and f2) or (not f2 and f1):
            try:
                raise PairedReadError(
                    "unpaired reads detected. Are files sorted? are "
                    "files of equal length?")
            except PairedReadError as e:
                raise PairedReadError(e).with_traceback(sys.exc_info()[2])
        else:
            assert f1.identifier.endswith("/1") and \
                f2.identifier.endswith("/2"), \
                "Reads in file 1 must end with /1 and reads in file 2 with /2"
            options.stdout.write(
                ">%s\n%s\n>%s\n%s\n" %
                (f1.identifier, f1.seq, f2.identifier, f2.seq))
            f1_count += 1

    E.info("output: %i pairs" % f1_count)

    # write footer and output benchmark information.
    E.Stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''
WrapperMuscle.py - 
======================================================

:Tags: Python

Code
----

'''
import os
import tempfile
import subprocess
import shutil

from types import *
from CGAT import Experiment as Experiment
from CGAT import Mali as Mali


class Error(Exception):

    """Base class for exceptions in this module."""

    def __str__(self):
        return str(self.message)


class ParsingError(Error):

    """Exception raised for errors while parsing

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message, line):
        self.message = message + " at line " + line


class UsageError(Error):

    """Exception raised for errors while starting

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message


class Muscle:

    mExecutable = "muscle -maxmb 10000"

    def __init__(self):

        self.mTempdir = None

    def __del__(self):

        if self.mTempdir and self.mTempdir != "tmp":
            os.system("rm -rf %s" % self.mTempdir)

    # ------------------------------------------------------------------------
    def Run(self, mali,
            tree=None,
            dump=0,
            test=False,
            options={}):

        self.mTempdir = tempfile.mkdtemp()
        self.mFilenameInput = "input"
        self.mFilenameOutput = "output"

        if test:
            print("# temporary directory is %s" % self.mTempdir)

        mali.writeToFile(open(self.mTempdir + "/" + self.mFilenameInput, "w"),
                         format="fasta")

        statement = " ".join((self.mExecutable,
                              "-in %s" % self.mFilenameInput,
                              "-out %s" % self.mFilenameOutput))

        s = subprocess.Popen(statement,
                             shell=True,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE,
                             cwd=self.mTempdir,
                             close_fds=True)

        (out, err) = s.communicate()

        if s.returncode != 0:
            raise UsageError("Error in running %s \n%s\n%s\nTemporary directory in %s" % (
                self.mExecutable, err, out, self.mTempdir))

        if dump:
            print("# stdout output of %s:\n%s\n######################################" % (
                self.mExecutable, out))

        result = Mali.Mali()

        result.readFromFile(
            open("%s/%s" % (self.mTempdir, self.mFilenameOutput), "r"), format="fasta")

        if not test:
            shutil.rmtree(self.mTempdir)

        return result

if __name__ == "__main__":

    parser = E.OptionParser(
        version="%prog version: $Id: WrapperMuscle.py 2784 2009-09-10 11:41:14Z andreas $")

    parser.add_option("-t", "--filename-input-tree", dest="filename_input_tree", type="string",
                      help="filename with tree information.")
    parser.add_option("-T", "--filename-output-tree", dest="filename_output_tree", type="string",
                      help="output filename with tree information.")
    parser.add_option("-p", "--program", dest="program", type="string",
                      help="program to use.")
    parser.add_option("-o", "--options", dest="options", type="string",
                      help="input options.")

    parser.set_defaults(
        filename_input_tree=None,
        filename_output_tree=None,
        program=None,
        options="",
    )

    (options, args) = Experiment.Start(parser)

    Experiment.Stop()



'''Bed2table.py - annotate intervals
=================================

:Tags: Genomics Intervals Summary

Purpose
-------

This script takes a bed-formatted file as input and annotates each interval.
Possible annotators are (see option '--counter'):

overlap

    compute overlap with intervals in other bed file. If the other bed
    file contains tracks, the overlap is computed per track.

peaks

    compute peak location in intervals. Requires one or more
    bam-files. This counter can also count within an secondary set of
    bam-files (--control-bam-file) and add this to the output.

composition-na

    compute nucleotide frequencies in intervals.

composition-cpg

    compute CpG densities and CpG observed / expected in intervals.

classifier-chipseq

   classify chipseq intervals. Requires a :term:`gff`
   file with genomic annotations (see :doc:`gtf2gff`.)

motif

   Search for a specified motif e.g. using --motif-sequence=TTTT.


Usage
-----

For example, the following command will compute the CpG composition
in the intervals supplied::

   cgat bed2table --counter=composition-cpg < in.bed > out.tsv

If the input is this :term:`bed` formatted file::

  chr19   60118   60120   1       100.0000
  chr19   60171   60173   2       100.0000
  chr19   60182   60184   3       100.0000
  chr19   60339   60341   4       100.0000
  chr19   60375   60377   5       100.0000
  chr19   60110   60118   noCpG   100.0000
  chr19   60118   60119   Conly   100.0000
  chr19   60119   60120   Gonly   100.0000

then the output is the following table:

+------+-----+-----+-----+--------+---------+-----------+----------+
|contig|start|end  |name |score   |CpG_count|CpG_density|CpG_ObsExp|
+------+-----+-----+-----+--------+---------+-----------+----------+
|chr19 |60118|60120|1    |100.0000|1        |1.0        |2.0       |
+------+-----+-----+-----+--------+---------+-----------+----------+
|chr19 |60171|60173|2    |100.0000|1        |1.0        |2.0       |
+------+-----+-----+-----+--------+---------+-----------+----------+
|chr19 |60182|60184|3    |100.0000|1        |1.0        |2.0       |
+------+-----+-----+-----+--------+---------+-----------+----------+
|chr19 |60339|60341|4    |100.0000|1        |1.0        |2.0       |
+------+-----+-----+-----+--------+---------+-----------+----------+
|chr19 |60375|60377|5    |100.0000|1        |1.0        |2.0       |
+------+-----+-----+-----+--------+---------+-----------+----------+
|chr19 |60110|60118|noCpG|100.0000|0        |0.0        |0.0       |
+------+-----+-----+-----+--------+---------+-----------+----------+
|chr19 |60118|60119|Conly|100.0000|0        |0.0        |0.0       |
+------+-----+-----+-----+--------+---------+-----------+----------+
|chr19 |60119|60120|Gonly|100.0000|0        |0.0        |0.0       |
+------+-----+-----+-----+--------+---------+-----------+----------+

Type::

   cgat bed2table.py --help

for command line help.

Command line options
---------------------

'''
import re
import sys
import collections
import CGAT.GTF as GTF
import CGAT.Bed as Bed
import CGAT.IOTools as IOTools
import CGAT.Experiment as E
import CGAT.IndexedFasta as IndexedFasta
import CGAT.SequenceProperties as SequenceProperties
import CGAT.Intervals as Intervals
import numpy
import pysam

import CGAT.GeneModelAnalysis as GeneModelAnalysis


class Counter(object):

    def __init__(self, fasta=None, *args, **kwargs):
        self.fasta = fasta

    def update(self, bed):
        self.bed = bed
        self.count(bed)

    def getSegments(self):
        return [self.bed]

    def getHeader(self):
        return '\t'.join(self.headers)


class CounterLength(Counter):
    headers = ['length']

    def count(self, bed):
        self.length = bed.end - bed.start

    def __str__(self):
        return str(self.length)


class CounterOverlap(Counter):

    '''count overlap for each interval in tracks.'''

    def __init__(self, filename, *args, **kwargs):

        assert filename is not None,\
            "please supply filename for CounterOverlap"

        Counter.__init__(self, *args, **kwargs)

        self.filename = filename

        E.info("reading intervals from %s" % self.filename)

        self.index = Bed.readAndIndex(
            IOTools.openFile(self.filename, "r"),
            per_track=True)

        E.info("read intervals for %s tracks" % len(self.index))

        self.tracks = list(self.index.keys())
        self.headers = []
        for track in self.tracks:
            self.headers.extend(["%s_nover" % track, "%s_bases" % track])

    def count(self, bed):
        '''update internal counts.'''

        results = []
        for track in self.tracks:
            try:
                overlaps = [(x[0], x[1])
                            for x in
                            self.index[track][bed.contig]
                            .find(bed.start, bed.end)]
            except KeyError:
                overlaps = []

            results.append((len(overlaps),
                            Intervals.calculateOverlap(
                                [(bed.start, bed.end), ],
                                Intervals.combine(overlaps))))

        self.data = results

    def __str__(self):
        '''output overlap of interval in *bed*'''

        r = []
        for track, result in zip(self.tracks, self.data):
            r.append("\t".join((str(result[0]), str(result[1]))))

        return "\t".join(r)

CounterPeaksResult = collections.namedtuple(
    "CounterPeaksResult", ("length nreads avgval peakval npeaks peakcenter"))


class CounterPeaks(Counter):

    '''compute number of extent of peaks in an interval.'''

    headers = None

    def __init__(self,
                 bamfiles,
                 offsets,
                 control_bamfiles,
                 control_offsets, *args, **kwargs):
        Counter.__init__(self, *args, **kwargs)
        if not bamfiles:
            raise ValueError("supply --bam-file options for readcoverage")

        assert len(offsets) == 0 or len(bamfiles) == len(
            offsets), "number of bamfiles not the same as number of offsets"
        assert len(control_offsets) == 0 or \
            len(control_bamfiles) == len(control_offsets), \
            "number of control bamfiles not the same as number of offsets"

        self.bamfiles = bamfiles
        self.offsets = offsets
        self.control_bamfiles = control_bamfiles
        self.control_offsets = control_offsets

        self.headers = list(CounterPeaksResult._fields)
        if self.control_bamfiles:
            self.headers.extend(
                ["control_%s" % x for x in CounterPeaksResult._fields])

    def _count(self, bed, bamfiles, offsets):
        '''count reads in bed interval.'''

        contig, start, end = bed.contig, bed.start, bed.end

        length = end - start
        try:
            counts = numpy.zeros(length)
        except ValueError as msg:
            raise ValueError("Error negative length obtained: "
                             " message=%s contig=%s, start=%s, end=%s" %
                             (msg, contig, start, end))
        nreads = 0

        if offsets:
            # if offsets are given, shift tags.
            for samfile, offset in zip(bamfiles, offsets):

                shift = offset // 2
                # for peak counting I follow the MACS protocoll,
                # see the function def __tags_call_peak in PeakDetect.py
                # In words
                # Only take the start of reads (taking into account the strand)
                # add d/2=offset to each side of peak and start accumulate
                # counts.
                # for counting, extend reads by offset
                # on + strand shift tags upstream
                # i.e. look at the downstream window
                xstart, xend = max(0, start - shift), max(0, end + shift)

                for read in samfile.fetch(contig, xstart, xend):
                    nreads += 1
                    # some reads are assigned to a contig and position, but
                    # are flagged as unmapped - these might not have an alen
                    # attribute.
                    if read.is_unmapped:
                        continue

                    if read.is_reverse:
                        # rstart = read.pos + read.alen - offset
                        # offset = 2 * shift
                        try:
                            rstart = read.pos + read.alen - offset
                        except TypeError as msg:
                            raise TypeError("Error message =", msg,
                                            "read.pos =", read.pos,
                                            "read.alen =", read.alen,
                                            "offset =", offset,
                                            "query name =", read.qname,
                                            "length of read =", read.rlen)
                    else:
                        rstart = read.pos + shift

                    rend = rstart + shift
                    rstart = max(0, rstart - start)
                    rend = min(length, rend - start)
                    counts[rstart:rend] += 1

        else:
            for samfile in bamfiles:
                for read in samfile.fetch(contig, start, end):
                    nreads += 1
                    rstart = max(0, read.pos - start)
                    rend = min(length, read.pos - start + read.rlen)
                    counts[rstart:rend] += 1

        length = end - start
        avgval = numpy.mean(counts)
        peakval = max(counts)

        # set other peak parameters
        peaks = numpy.array(list(range(0, length)))[counts >= peakval]
        npeaks = len(peaks)
        # peakcenter is median coordinate between peaks
        # such that it is a valid peak in the middle
        peakcenter = start + peaks[npeaks // 2]

        return CounterPeaksResult(length, nreads, avgval,
                                  peakval, npeaks, peakcenter)

    def count(self, bed):
        '''count reads per position.

        If offsets are given, shift tags by offset / 2 and extend
        by offset / 2.
        '''

        self.result = self._count(bed, self.bamfiles, self.offsets)
        if self.control_bamfiles:
            self.control = self._count(
                bed, self.control_bamfiles, self.control_offsets)

    def __str__(self):
        if self.control_bamfiles:
            return "\t".join(map(str, self.result + self.control))
        else:
            return "\t".join(map(str, self.result))


class CounterCompositionNucleotides(Counter):

    headers = SequenceProperties.SequencePropertiesNA().getHeaders()

    def __init__(self, *args, **kwargs):
        Counter.__init__(self, *args, **kwargs)
        self.result_class = SequenceProperties.SequencePropertiesNA
        assert self.fasta, "Counter requires a genomic sequence"

    def count(self, bed):
        s = self.fasta.getSequence(bed.contig, "+", bed.start, bed.end)
        self.result = self.result_class()
        self.result.loadSequence(s)

    def __str__(self):
        return str(self.result)


class CounterMotif(Counter):

    headers = ['motif_counts']

    def __init__(self, motif, *args, **kwargs):
        Counter.__init__(self, *args, **kwargs)
        self.motif = motif

    def count(self, bed):
        s = self.fasta.getSequence(bed.contig, "+", bed.start, bed.end)
        self.result = len([x for x in
                           re.finditer(r'(?=(%s))' % self.motif, s)])

    def __str__(self):
        return str(self.result)


class CounterCompositionCpG(CounterCompositionNucleotides):

    '''compute CpG frequencies as well as nucleotide frequencies.

    Note that CpG density is calculated across the merged exons of a
    transcript. Thus, there might be difference between the CpG on a
    genomic level and on the transrcipt level depending on how many
    genomic CpG are lost across an intron-exon boundary or how many
    transcript CpG are created by exon fusion.

    '''

    headers = SequenceProperties.SequencePropertiesCpg().getHeaders()

    def __init__(self, *args, **kwargs):
        CounterCompositionNucleotides.__init__(self, *args, **kwargs)
        self.result_class = SequenceProperties.SequencePropertiesCpg

    def count(self, bed):

        s = self.fasta.getSequence(bed.contig, "+", bed.start, bed.end)
        self.result = self.result_class()
        self.result.loadSequence(s)


class ClassifierChIPSeq(GeneModelAnalysis.Classifier):

    """classify ChIPSeq intervals based on a reference annotation.

    This assumes the input is a genome annotation derived from an
    ENSEMBL gtf file created with gff2gtf.py.

    In contrast to transcripts, the intervals are fuzzy. Hence the
    classification is based on a mixture of full/partial overlap.

    An interval is classified as:

    cds
       mostly part of a CDS. These would be intervals fully within a CDS exon.
    utr
       mostly part of UTR. These are intervals fully within the UTR of a gene.
    intergenic
       mostly intergenic. These are intervals fully within the
       intergenic region and more than 1kb from the closest exon.
    upstream
       not any of the above and partly upstream of a gene. These are
       intervals that might overlap part of the UTR or the 1kb segment
       before to the 5'-terminal exon of a gene.
    downstream
       not any of the abore and partly downstream of a gene. These are
       intervals that might overlap part of the UTR or the 1kb segment
       after to the 3'-terminal exon of a gene.
    intronic
       not any of the above and partly intronic. Note that these could
       also include promotors of short alternative transcripts that
       skip one or more of the first exons.
    ambiguous
       none of the above

    """

    header = ["is_cds", "is_utr", "is_upstream", "is_downstream",
              "is_intronic", "is_intergenic", "is_flank", "is_ambiguous"]

    # sources to use for classification
    sources = ("", )

    # minimum coverage of a transcript to assign it to a class
    mThresholdMinCoverage = 95

    # full coverage of a transcript to assign it to a class
    mThresholdFullCoverage = 99

    # some coverage of a transcript to assign it to a class
    mThresholdSomeCoverage = 10

    def update(self, bed):

        # convert to a gtf entry
        gtf = GTF.Entry()

        gtf.fromBed(bed)
        gtf.feature = 'exon'
        GeneModelAnalysis.Classifier.update(self, [gtf])

    def count(self):

        for key in self.mKeys:
            self.mCounters[key].update(self.mGFFs)

        def s_min(*args):
            return sum([abs(self.mCounters[x].mPOverlap1)
                        for x in args]) >= self.mThresholdMinCoverage

        def s_excl(*args):
            return sum([abs(self.mCounters[x].mPOverlap1)
                        for x in args]) < (100 - self.mThresholdMinCoverage)

        def s_full(*args):
            return sum([abs(self.mCounters[x].mPOverlap1)
                        for x in args]) >= self.mThresholdFullCoverage

        def s_some(*args):
            return sum([abs(self.mCounters[x].mPOverlap1)
                        for x in args]) >= self.mThresholdSomeCoverage

        self.mIsCDS, self.mIsUTR, self.mIsIntergenic = False, False, False
        self.mIsUpStream = False
        self.mIsDownStream = False
        self.mIsIntronic = False
        self.mIsFlank, self.mIsAmbiguous = False, False

        self.mIsCDS = s_full(":CDS")
        self.mIsUTR = s_full(":UTR", ":UTR3", ":UTR5")
        self.mIsIntergenic = s_full(":intergenic", ":telomeric")

        if not(self.mIsCDS or self.mIsUTR or self.mIsIntergenic):
            self.mIsUpStream = s_some(":5flank", ":UTR5")
            if not self.mIsUpStream:
                self.mIsDownStream = s_some(":3flank", ":UTR3")
                if not self.mIsDownStream:
                    self.mIsIntronic = s_some(":intronic")
                    if not self.mIsIntronic:
                        self.mIsFlank = s_some(":flank")

        self.mIsAmbiguous = not(self.mIsUTR or
                                self.mIsIntergenic or
                                self.mIsIntronic or
                                self.mIsCDS or
                                self.mIsUpStream or
                                self.mIsDownStream or
                                self.mIsFlank)

    def __str__(self):

        def to(v):
            if v:
                return "1"
            else:
                return "0"

        h = [to(x) for x in (self.mIsCDS,
                             self.mIsUTR,
                             self.mIsUpStream,
                             self.mIsDownStream,
                             self.mIsIntronic,
                             self.mIsIntergenic,
                             self.mIsFlank,
                             self.mIsAmbiguous,
                             )]

        for key in self.mKeys:
            h.append(str(self.mCounters[key]))
        return "\t".join(h)


def main(argv=None):

    if argv is None:
        argv = sys.argv

    parser = E.OptionParser(
        version="%prog version: $Id$",
        usage=globals()["__doc__"])

    parser.add_option("-g", "--genome-file", dest="genome_file", type="string",
                      help="filename with genome [default=%default].")

    parser.add_option(
        "-b", "--bam-file", dest="bam_files", type="string",
        help="filename with read mapping information. Multiple files can be "
        "submitted in a comma-separated list [default=%default].")

    parser.add_option(
        "--control-bam-file", dest="control_bam_files", type="string",
        help="filename with read mapping information for input/control. "
        "Multiple files can be submitted in a comma-separated list "
        "[default=%default].")

    parser.add_option(
        "--filename-format", dest="filename_format", type="choice",
        choices=("bed", "gff", "gtf"),
        help="format of secondary stream [default=%default].")

    parser.add_option(
        "-c", "--counter", dest="counters", type="choice", action="append",
        choices=("length",
                 "overlap",
                 "peaks",
                 "composition-na",
                 "composition-cpg",
                 "classifier-chipseq",
                 "motif"),
        help="select counters to apply [default=%default].")

    parser.add_option(
        "--motif-sequence", dest="motif_sequence", type="string",
        help="specify a sequence to search for"
        "[default=%default].")

    parser.add_option(
        "-o", "--offset", dest="offsets", type="int", action="append",
        help="tag offsets for tag counting - supply as many as there "
        "are bam-files [default=%default].")

    parser.add_option(
        "--control-offset", dest="control_offsets", type="int",
        action="append",
        help="control tag offsets for tag counting - supply as many as "
        "there are bam-files [default=%default].")

    parser.add_option(
        "-a", "--output-all-fields", dest="all_fields", action="store_true",
        help="output all fields in original bed file, by default only "
        "the first 4 are output [default=%default].")

    parser.add_option(
        "--output-bed-headers", dest="bed_headers", type="string",
        help="supply ',' separated list of headers for bed component "
        "[default=%default].")

    parser.add_option(
        "-f", "--gff-file", dest="filename_gff", type="string",
        action="append", metavar='bed',
        help="filename with extra gff files. The order is important "
        "[default=%default].")

    parser.add_option(
        "--has-header", dest="has_header", action="store_true",
        help="bed file with headers. Headers and first columns are "
        "preserved [default=%default]")

    parser.set_defaults(
        genome_file=None,
        counters=[],
        bam_files=None,
        offsets=[],
        control_bam_files=None,
        control_offsets=[],
        all_fields=False,
        filename_format=None,
        bed_headers=None,
        filename_gff=[],
        has_header=False,
        motif_sequence=None
    )

    (options, args) = E.Start(parser)

    if options.bed_headers is not None:
        bed_headers = [x.strip() for x in options.bed_headers.split(",")]
        if len(bed_headers) < 3:
            raise ValueError("a bed file needs at least three columns")
    else:
        bed_headers = None

    if options.has_header:
        while 1:
            line = options.stdin.readline()
            if not line:
                E.warn("empty bed file with no header")
                E.Stop()
                return
            if not line.startswith("#"):
                break
        bed_headers = line[:-1].split("\t")

    if "motif" in options.counters and not options.motif_sequence:
        raise ValueError("if using motif must specify a motif-sequence")

    # get files
    if options.genome_file:
        fasta = IndexedFasta.IndexedFasta(options.genome_file)
    else:
        fasta = None

    if options.bam_files:
        bam_files = []
        for bamfile in options.bam_files.split(","):
            bam_files.append(pysam.AlignmentFile(bamfile, "rb"))
    else:
        bam_files = None

    if options.control_bam_files:
        control_bam_files = []
        for bamfile in options.control_bam_files.split(","):
            control_bam_files.append(pysam.AlignmentFile(bamfile, "rb"))
    else:
        control_bam_files = None

    counters = []

    for c in options.counters:
        if c == "length":
            counters.append(CounterLength(fasta=fasta,
                                          options=options))

        elif c == "overlap":
            counters.append(CounterOverlap(filename=options.filename_gff[0],
                                           fasta=fasta,
                                           options=options))
            del options.filename_gff[0]
        elif c == "peaks":
            counters.append(CounterPeaks(bam_files,
                                         options.offsets,
                                         control_bam_files,
                                         options.control_offsets,
                                         options=options))
        elif c == "composition-na":
            counters.append(CounterCompositionNucleotides(fasta=fasta,
                                                          options=options))
        elif c == "composition-cpg":
            counters.append(CounterCompositionCpG(fasta=fasta,
                                                  options=options))
        elif c == "classifier-chipseq":
            counters.append(ClassifierChIPSeq(
                filename_gff=options.filename_gff,
                fasta=fasta,
                options=options,
                prefix=None))
            del options.filename_gff[0]

        elif c == "motif":
            counters.append(CounterMotif(fasta=fasta,
                                         motif=options.motif_sequence))

    extra_fields = None

    for bed in Bed.iterator(options.stdin):

        if extra_fields is None:

            # output explicitely given headers
            if bed_headers:
                if len(bed_headers) > bed.columns:
                    raise ValueError(
                        "insufficient columns (%i, expected %i) in %s" %
                        (bed.columns, len(bed_headers), str(bed)))

            else:
                bed_headers = Bed.Headers[:bed.columns]

            options.stdout.write("\t".join(bed_headers))
            options.stdout.write("\t" + "\t".join(
                [x.getHeader() for x in counters]) + "\n")

            extra_fields = list(range(len(bed_headers) - 3))

        for counter in counters:
            counter.update(bed)

        if options.all_fields:
            options.stdout.write(str(bed))
        else:
            options.stdout.write(
                "\t".join([bed.contig,
                           str(bed.start),
                           str(bed.end)] + [bed.fields[x]
                                            for x in extra_fields]))
        for counter in counters:
            options.stdout.write("\t%s" % str(counter))

        options.stdout.write("\n")

    E.Stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



"""SequencePairProperties.py - Computing metrics for aligned sequences
======================================================================

This module provides methods for extracting and reporting sequence
properties of aligned nucleotide sequences such as percent identity,
substitution rate, etc. Usage is the same as
:mod:`SequencePairProperties`.


Reference
---------

"""
import numpy
from CGAT import Mali as Mali
from CGAT import WrapperCodeML as WrapperCodeML


class SequencePairProperties:

    mPseudoCounts = 1

    def __init__(self):

        self.mLength = 0
        self.mNCodons = 0

    def addProperties(self, other):
        """add properties."""

        self.mLength += other.mLength
        self.mNCodons += other.mNCodons

    def updateProperties(self):
        pass

    def loadPair(self, seq1, seq2):
        """load sequence properties from a pair."""

    def __str__(self):
        return "\t".join(self.getFields())

    def __call__(self, seq1, seq2):
        self.loadPair(seq1.upper(), seq2.upper())

    def getFields(self):
        self.updateProperties()
        return []

    def getHeaders(self):
        return []

    def getHeader(self):
        return "\t".join(self.getHeaders)


class SequencePairPropertiesDistance(SequencePairProperties):

    """base class for distance estimators."""

    def __init__(self, *args, **kwargs):
        SequencePairProperties.__init__(self, *args, **kwargs)


class SequencePairPropertiesBaseML (SequencePairPropertiesDistance):

    """Counts for nucleic acid sequences.

    The first characters are ACGT.
    """

    mFormat = "%6.4f"

    def __init__(self, options, *args, **kwargs):
        SequencePairPropertiesDistance.__init__(self, *args, **kwargs)

        self.mBaseml = WrapperCodeML.BaseML()
        self.mBaseml.SetOptions(options)

        if options.loglevel >= 3:
            self.mDump = True
            self.mTest = True
        else:
            self.mDump = False
            self.mTest = False

    def loadPair(self, seq1, seq2):

        temp_mali = Mali.Mali()
        temp_mali.addSequence("seq1", 0, len(seq1), seq1)
        temp_mali.addSequence("seq2", 0, len(seq2), seq2)

        try:
            self.mResult = self.mBaseml.Run(temp_mali,
                                            tree="(seq1,seq2);",
                                            dump=self.mDump,
                                            test=self.mTest)
        except WrapperCodeML.UsageError:
            self.mResult = None

    def getHeaders(self):
        return ["distance", "lnl", "kappa", "converged"]

    def __str__(self):

        if not self.mResult:
            return "\t".join(["na"] * 4)

        if self.mResult.mKappa != "na":
            o_kappa = self.mFormat % self.mResult.mKappa
        else:
            o_kappa = self.mResult.mKappa

        if self.mResult.mCheckConvergence:
            o_converged = "0"
        else:
            o_converged = "1"

        return "\t".join((
            self.mFormat % self.mResult.mDistanceMatrix["seq1"]["seq2"],
            self.mFormat % self.mResult.mLogLikelihood,
            o_kappa,
            o_converged))


class SequencePairPropertiesCountsNa (SequencePairProperties):

    """Counts for nucleic acid sequences.

    The first characters are ACGT.
    """

    mAlphabet = "ACGT"
    mGapChar = "-"

    def __init__(self, *args, **kwargs):
        SequencePairProperties.__init__(self, *args, **kwargs)
        self.mNIdentical = 0
        self.mNAligned = 0
        self.mNDifferent = 0
        self.mNTransitions = 0
        self.mNTransversions = 0
        self.mNUnaligned1 = 0
        self.mNUnaligned2 = 0
        self.mMapChar2Pos = {}
        self.mPercentGC = 0
        self.mMatrix = []

    def buildSubstitutionMatrix(self, seq1, seq2, alphabet):
        """given a pair of sequences, calculate
        a substitution matrix for the given alphabet.
        """
        if len(seq1) != len(seq2):
            raise ValueError("two sequences of unequal length submitted")

        nchars = len(alphabet)
        matrix = numpy.zeros((nchars, nchars), numpy.int)
        map_char2pos = {}
        for x in alphabet:
            map_char2pos[x] = len(map_char2pos)

        for x in range(len(seq1)):
            try:
                matrix[map_char2pos[seq1[x]], map_char2pos[seq2[x]]] += 1
            except KeyError:
                continue

        return matrix

    def loadPair(self, seq1, seq2):
        """load sequence properties from a pair.
        """
        SequencePairProperties.loadPair(self, seq1, seq2)

        alphabet = self.mAlphabet + self.mGapChar

        map_char2pos = {}
        for x in alphabet:
            map_char2pos[x] = len(map_char2pos)

        # build coordinates for various substitution subsets
        transitions, transversions = [], []
        for x in ("AG", "GA", "CT", "TC"):
            transitions.append((map_char2pos[x[0]], map_char2pos[x[1]]))

        for x in ("AT", "TA", "GT", "TG", "GC", "CG", "AC", "CA"):
            transversions.append((map_char2pos[x[0]], map_char2pos[x[1]]))

        matrix = self.buildSubstitutionMatrix(seq1, seq2, alphabet)
        matrix_acgt = matrix[0:4, 0:4]

        self.mMatrix = matrix
        self.mMapChar2Pos = map_char2pos
        self.mNAligned = numpy.sum(numpy.ravel(matrix_acgt))
        self.mNIdentical = numpy.sum(numpy.trace(matrix_acgt))
        self.mNTransitions = numpy.sum([matrix[x] for x in transitions])
        self.mNTransversions = numpy.sum([matrix[x] for x in transversions])
        self.mNDifferent = self.mNAligned - self.mNIdentical
        self.mNUnaligned1 = numpy.sum(numpy.ravel(matrix[0:4, 4]))
        self.mNUnaligned2 = numpy.sum(numpy.ravel(matrix[4, 0:4]))

        cp = self.mMapChar2Pos['C']
        gp = self.mMapChar2Pos['G']

        # sum all rows and columns that have a least one G or C
        # and remove those that have two in order to not double count
        gc = numpy.sum(
            self.mMatrix[0:4, cp] +
            self.mMatrix[0:4, gp] +
            self.mMatrix[cp, 0:4] +
            self.mMatrix[gp, 0:4]) \
            - self.mMatrix[cp, cp] \
            - self.mMatrix[gp, gp] \
            - self.mMatrix[cp, gp] \
            - self.mMatrix[gp, cp]

        try:
            self.mPercentGC = "%5.2f" % (100.0 * float(gc) / self.mNAligned)
        except ZeroDivisionError:
            self.mPercentGC = "na"

    def __str__(self):

        return "\t".join(map(str, (self.mNIdentical, self.mNAligned,
                                   self.mNDifferent,
                                   self.mNTransitions, self.mNTransversions,
                                   self.mNUnaligned1, self.mNUnaligned2,
                                   self.mPercentGC)))

    def getHeaders(self):
        return ["identical", "aligned", "different", "transitions",
                "transversions", "unaligned1", "unaligned2", "pgc"]


class SequencePairPropertiesCountsCodons(SequencePairPropertiesCountsNa):

    """the first characters are ACGT."""

    def __init__(self):
        SequencePairPropertiesCountsNa.__init__(self)
        self.mNNonSynonymous = 0
        self.mNSynonymous = 0

    def __str__(self):

        return "\t".join((SequencePairPropertiesCountsNa.__str__(self),
                          str(self.mNNonSynonymous),
                          str(self.mNSynonymous)))

    def getHeaders(self):
        return ["nonsyn", "nsyn"]


class SequencePairPropertiesPID(SequencePairPropertiesDistance):
    """Percent identity.

    The percent identity is the ratio of the number of identical
    residues divided by the number of aligned residues.

    """
    mGapChars = ("-", "."),

    def __init__(self, *args, **kwargs):
        SequencePairPropertiesDistance.__init__(self, *args, **kwargs)
        self.mPID = 0

    def loadPair(self, seq1, seq2):

        nidentical, naligned = 0, 0
        for c1, c2 in zip(seq1.upper(), seq2.upper()):
            if c1 in self.mGapChars or c2 in self.mGapChars:
                continue
            naligned += 1
            if c1 == c2:
                nidentical += 1

        self.mPID = 100.0 * float(nidentical) / naligned

    def getHeaders(self):
        return ["distance", ]

    def __str__(self):
        return self.mFormat % self.mPID



"""gtf2gff.py - convert a transcript set to genomic features
=========================================================

:Tags: Genomics Genesets Intervals Transformation GTF GFF

Purpose
-------

This scripts converts a transcript set in a :term:`gtf` formatted file
into a set of features in a :term:`gff` formatted file.

In other words, a gene set (gtf), which constitutes a hierarchical set
of annotations, will be converted into a non-hierarchical list of
genomic segments.

Various methods can be used to do the conversion (see command line
argument ``--method``):

exons
   annotate exons. Exonic segments are classified according to the
   transcript structure.

genome/full
   annotate genome with gene set. Genomic segments are labeled
   ``intronic``, ``intergenic``, etc. This annotation aggregates the
   information of multiple genes such that each annotation is either
   valid or ambiguous.

genes
    annotate genome using the information on a gene-by-gene basis.
    Multiple overlapping annotations will be created for each transcript.
    Redundant annotations will be merged.

great-domains
   regulatory domains using the basal+extended model according to GREAT.

promotors
   declare promoter regions. These segments might be overlapping. A promotor
   is the region x kb upstream of a transcription start site. The option
   ``--promotor-size`` sets the region width.

regulons
   declare regulatory regions. Regulatory regions contain the region x
   kb of upstream and downstream of a transciption start site. The
   options ``--upstream-extension`` and ``-downstream`` set the region width.

tts-regulons
   declare tts regulatory regions. tts-regulatory regions contain the
   region x kb of upstream and downstream of a transciption
   termination site. The options ``--upstream-extension`` and ``-downstream``
   set the region width.

territories
   build gene territories around full length genes.

tss-territories
   build gene territories around transcription start sites.

In a simple setting, assume we have the two genes below, the first
with a single transcript on the positive strand, the second on the
negative strand::

          Gene A                    Gene B
           |---|                 |---|  |---|
        >>>>   >>>>           <<<<   <<<<   <<<<

   Genome (simplified result without UTRs and flanks)

          exon   exon         exon   exon   exon
   ..---><--><-><--><---------<--><-><--><-><--><-----...
   intergenic intron  intergenic  intron intron intergenic

   Territories

        Gene A                    Gene B
   <---------------------><------------------------------>

   TSS-Territories

        Gene A                    Gene B
   <-------->            <----------->

   Promotors

   <---->              <---->

Genome
++++++

If ``--method=genome``, the gene set is used to annotate the complete genome.

.. note::
   The gtf file has to be sorted first by contig and then by position.

A segment in the genome will either be covered by:

cds
   a coding exon (also: CDS, start_codon).

utr
   a UTR (also: stop_codon)

5flank, 3flank, flank
   an upstream/downstream segment of defined size. If the intergenic
   region is too small to accomodate a flank, the regions is just
   'flank'.

intergenic
   intergenic region.

5telomeric, 3telomeric
   telomeric region (before/after first/last gene).

intronic
   intronic region. An intron has a minimum size of 30 bases.

frameshift
   frameshift. Introns of less than 4 residues length

ambiguous
   in case of overlapping genes, regions are designated ambiguous

unknown
   unknown are ``intronic`` regions that are less than the
   minimum size of an intron (default: 30) and larger than the size of
   frameshift (default:4).  These could be either genuine small
   introns or they could be artefactual arising from collapsing the
   exons within a gene model.

All segments are annotated by their closest gene. Intergenic regions are
annotated with their two neighbouring genes. The upstream gene is listed
in the attribute gene_id, the downstream one is listed in the attribute
downstream_gene_id.

Genes
+++++

If ``--method=genes``, the gene set is used to annotate the complete genome.

.. note::
   The gtf file has to be sorted by gene.

A segment in the genome will be annotated as:

cds
   a coding exon

utr5, utr3
   a 5' or 3' utr

exon
   an exon. Exons are further classified into first, middle and last exons.

intronic
   an intronic region. Intronic regions are further divided into
   first, middle, last.

upstream, downstream
   upstream/downstream regions in 5 intervals of a total of 1kb (see
   option --flank-size to increase the total size).

.. _territories:

Territories
+++++++++++

If ``--method=territories``, the gene set is used to define gene
territories.  Territories are segments around genes and are
non-overlapping. Exons in a gene are merged and the resulting the
region is enlarged by --radius. Overlapping territories are divided at
the midpoint between the two genes. The maximum extent of a territory
is limited by the option ``--territory-extension``

.. note::
   The gtf file has to be sorted first by contig and then by position.

.. note::
   Genes should already have been merged (gtf2gtf --merge-transcripts)

TSSTerritories
++++++++++++++

If ``--method=tss-territories``, the gene set is used to define gene
territories.  Instead of the full gene length as in
:ref:`territories`, only the tss is used to define a
territory. Territories are segments around genes and are
non-overlapping.  Overlapping territories are divided at the midpoint
between the two genes. The maximum extent of a territory is limited by
the option ``--territory-extension``.

.. note::
   The gtf file has to be sorted first by contig and then by position.

.. note::
   Genes should already have been merged (gtf2gtf --merge-transcripts)

The domain definitions corresponds to the ``nearest gene`` rule in GREAT.

GREAT-Domains
+++++++++++++

Define GREAT regulatory domains. Each TSS in a gene is associated with
a basal region. The basal region is then extended upstream to the
basal region of the closest gene, but at most to --radius. In the case
of overlapping genes, the extension is towards the next
non-overlapping gene.

This is the "basal plus extension" rule in GREAT. Commonly used are
5+1 with 1 Mb extension.  To achieve this, use for example::

   cgat gtf2gff \
   --genome-file=hg19 \
   --method=great-domains \
   --upstream-extension=5000 \
   --downstream-extension=1000 \
   --territory-extension=1000000 \
   < in.gtf > out.gff

If there are a multiple TSS in a transcript, the basal region extends from the
first to the last TSS plus the upstream/downstream flank.

Exons
+++++

If ``--method=exons``, exons are annotated by their dispensibility.

.. note::
   The gtf file should be sorted by genes

For each exon, the following additional fields are added to the gtf file:

ntranscripts
   number of transcripts
nused
   number of transcripts using this exon
positions
   positions of exon within transcripts. This is a ``,`` separated
   list of tuples ``pos:total``. For example, ``1:10,5:8`` indicates
   an exon that appears in first position in a ten exon transcript and
   fifth position in an eight exon transcript. The position is
   according to the direction of transcription.

.. note::
   overlapping but non-identical exons, for example due to internal
   splice sites, are listed as separate exons. Thus the output is not
   fully flat as some segments could be overlapping (see output
   variable ``noverlapping`` in the log file).

The following example uses an ENSEMBL gene set:: (needs genome-file to
run)

   gunzip < Mus_musculus.NCBIM37.55.gtf.gz | awk '$3 == "CDS"' | python gtf2gff.py --method=exons --restrict-source=protein_coding

Promoters
+++++++++

If ``--method=promotors``, putative promotor regions are output. A
promoter is a pre-defined segment upstream of the transcription start
site. As the actual start site is usually not known, the start of the
first exon within a transcript is used as a proxy. A gene can have
several promotors associated with it, but overlapping promotor regions
of the same gene will be merged. A promoter can extend into an
adjacent upstream gene.

The ``--restrict-source`` option determines which GTF entries are
output. The default is to output all entries but the user can choose
from protein_coding, pseudogene or lncRNA.

The size of the promotor region can be specified by the command line
argument ``--promotor-size``.

Regulons
+++++++++

If ``--method=regulons``, putative regulon regions are output. This is similar
to a ``promotor``, but the region extends both upstream and downstream from
the transcription start site.

The ``--restrict-source`` option determines which GTF entries are
output. The default is to output all entries but the user can choose
from protein_coding, pseudogene or lncRNA.

The size of the promotor region can be specified by the command line
argument ``--upstream-extension`` and ``--downstream-extension``

If ``--method=tts-regulons``, regulons will be defined around the
transcription termination site.

Usage
-----

Type::

    cgat gtf2gff --method=genome --genome-file=hg19 < geneset.gtf > annotations.gff

For command line help::

    cgat gtf2gff --help

Command line options
---------------------

"""

import sys
import collections
import itertools

import CGAT.GTF as GTF
import CGAT.Experiment as E
import CGAT.IndexedFasta as IndexedFasta
import CGAT.Genomics as Genomics
import CGAT.Intervals as Intervals
import CGAT.IOTools as IOTools


def addSegment(feature, start, end, template, options):
    """add a generic segment of type *feature*.
    """
    if start >= end:
        return 0

    entry = GTF.Entry()

    if isinstance(template, tuple):
        entry.copy(template[0])
        entry.clearAttributes()
        entry.addAttribute("downstream_gene_id", template[1].gene_id)
    else:
        entry.copy(template)
        entry.clearAttributes()

    entry.start, entry.end = start, end
    entry.feature = feature
    if feature not in ("exon", "CDS", "UTR", "UTR3", "UTR5"):
        entry.score = "."
    options.stdout.write(str(entry) + "\n")

    return 1


def addFlank(start, end, template, options):
    """add a flank.
    """
    is_positive = Genomics.IsPositiveStrand(template.strand)
    is_before = end <= template.start
    if (is_before and is_positive) or (not is_before and not is_positive):
        name = "5flank"
    else:
        name = "3flank"

    return addSegment(name, start, end, template, options)


def addIntergenicSegment(last, this, fasta, options):
    """add an intergenic segment between last and this.

    At telomeres, either can be None.
    """
    if not this and not last:
        return 0

    nadded = 0
    if not this:
        # last telomere
        try:
            lcontig = fasta.getLength(last.contig)
        except KeyError as msg:
            if options.ignore_missing:
                return nadded
            else:
                raise KeyError(msg)
        flank = min(last.end + options.flank, lcontig)
        nadded += addFlank(last.end, flank, last, options)
        nadded += addSegment("telomeric", flank, lcontig, last, options)
    elif not last:
        # first telomere
        flank = max(0, this.start - options.flank)
        nadded += addSegment("telomeric", 0, flank, this, options)
        nadded += addFlank(flank, this.start, this, options)
    else:
        # intergenic region
        d = this.start - last.end
        flank = options.flank
        if d > flank * 2:
            nadded += addFlank(last.end, last.end + flank, last, options)
            nadded += addSegment("intergenic", last.end +
                                 flank, this.start - flank,
                                 (last, this), options)
            nadded += addFlank(this.start - flank, this.start, this, options)
        else:
            # add short flank between two genes. If they can not agree
            # on the directionality, "flank" is used.
            is_positive1 = Genomics.IsPositiveStrand(last.strand)
            is_positive2 = Genomics.IsPositiveStrand(this.strand)
            if is_positive1 and not is_positive2:
                key = "3flank"
            elif not is_positive1 and is_positive2:
                key = "5flank"
            else:
                key = "flank"
            nadded += addSegment(key, last.end, this.start,
                                 (last, this), options)

    return nadded


def buildTerritories(iterator, fasta, method, options):
    """build gene territories.

    Exons in a gene are merged and the resulting segments enlarged by
    --radius. Territories overlapping are divided in the midpoint
    between the two genes.

    If *method* is ``gene``, gene territories will be built.
    If *method* is ``tss``, tss territories will be built.

    """

    ninput, noutput, nambiguous = 0, 0, 0

    assert method in ("gene", "tss")

    dr = 2 * options.radius

    prev_pos = 0
    last_contig = None
    gff = None

    def _iterator(iterator):
        """yield gene plus the locations of the end of the previous gene and
        start of next gene"""

        last_end, prev_end = 0, 0
        last_contig = None
        last = None
        for matches in GTF.iterator_overlaps(iterator):

            this_start = min([x.start for x in matches])
            this_end = max([x.end for x in matches])

            if method == "tss":
                # restrict to tss
                if matches[0].strand == "+":
                    this_end = this_start + 1
                else:
                    this_start = this_end - 1

            this_contig = matches[0].contig

            if last_contig != this_contig:
                if last:
                    yield prev_end, last, fasta.getLength(last_contig)
                last_end, prev_end = 0, 0
            else:
                yield prev_end, last, this_start

            prev_end = last_end
            last_end = this_end
            last = matches
            last_contig = this_contig

        if last:
            yield prev_end, last, fasta.getLength(last_contig)

    for last_end, matches, next_start in _iterator(iterator):

        gff = GTF.Entry().copy(matches[0])

        start = min([x.start for x in matches])
        end = max([x.end for x in matches])

        if method == "tss":
            # restrict to tss
            if matches[0].strand == "+":
                end = start + 1
            else:
                start = end - 1

        d = start - last_end
        if d < dr:
            start -= d // 2
        else:
            start -= options.radius

        d = next_start - end
        if d < dr:
            end += d // 2
        else:
            end += options.radius

        gff.gene_id = ":".join(sorted(set([x.gene_id for x in matches])))
        gff.transcript_id = gff.gene_id
        gff.start, gff.end = start, end

        nsegments = len(matches)
        if nsegments > 1:
            gff.addAttribute("ambiguous", nsegments)
            nambiguous += 1

        assert gff.start < gff.end, "invalid segment: %s" % str(gff)
        options.stdout.write(str(gff) + "\n")
        noutput += 1

    E.info("ninput=%i, noutput=%i, nambiguous=%i" %
           (ninput, noutput, nambiguous))


def annotateGenome(iterator, fasta, options):
    """perform a full segmentation of the genome (UTR, exon, intron ...)
    """

    ninput, noutput, nadded, nambiguous, nframeshifts, nunknown = 0, 0, 0, 0, 0, 0
    last = None
    is_ambiguous = False

    for this in iterator:
        ninput += 1

        E.debug("last=%s" % str(last))
        E.debug("this=%s" % str(this))
        E.debug("is_ambiguous=%s" % str(is_ambiguous))

        if last and last.contig == this.contig:
            # check if file is sorted correctly
            assert last.start <= this.start, "input file needs to be sorted by contig, start"
            if last.end <= this.start:
                if not is_ambiguous:
                    if last.gene_id != this.gene_id:
                        nadded += addIntergenicSegment(last,
                                                       this, fasta, options)
                    else:
                        d = this.start - last.end
                        if d >= options.min_intron_length:
                            nadded += addSegment("intronic",
                                                 last.end,
                                                 this.start,
                                                 last,
                                                 options)
                        elif d <= options.max_frameshift_length:
                            nframeshifts += addSegment("frameshift",
                                                       last.end,
                                                       this.start,
                                                       last,
                                                       options)
                        else:
                            nunknown += addSegment("unknown",
                                                   last.end,
                                                   this.start,
                                                   last,
                                                   options)
                else:
                    if last.feature == this.feature and \
                       last.gene_id == this.gene_id:
                        nambiguous += addSegment(
                            last.feature,
                            last.end, this.start,
                            last, options)
                    else:
                        nambiguous += addSegment(
                            "ambiguous",
                            last.end, this.start,
                            last, options)
                    is_ambiguous = False
                last = this
            elif last.end > this.start:
                if last.gene_id != this.gene_id:
                    # flag next region as ambiguous
                    is_ambiguous = True
                last.end = this.end
        else:
            nadded += addIntergenicSegment(last, None, fasta, options)
            nadded += addIntergenicSegment(None, this, fasta, options)
            last = this

        options.stdout.write("%s\n" % str(this))
        noutput += 1

    E.info(
        "ninput=%i, noutput=%i, nadded=%i, nambiguous=%i, nframeshifts=%i, nunknown=%i" %
        (ninput, noutput, nadded, nambiguous, nframeshifts, nunknown))


def annotateExons(iterator, fasta, options):
    """annotate exons within iterator."""

    gene_iterator = GTF.gene_iterator(iterator)

    ninput, noutput, noverlapping = 0, 0, 0

    for this in gene_iterator:
        ninput += 1
        intervals = collections.defaultdict(list)
        ntranscripts = len(this)

        is_negative_strand = Genomics.IsNegativeStrand(this[0][0].strand)

        for exons in this:
            # make sure these are sorted correctly
            exons.sort(key=lambda x: x.start)
            if is_negative_strand:
                exons.reverse()

            nexons = len(exons)
            for i, e in enumerate(exons):
                intervals[(e.start, e.end)].append((i + 1, nexons))

        gtf = GTF.Entry()
        gtf.fromGTF(this[0][0], this[0][0].gene_id, this[0][0].gene_id)
        gtf.addAttribute("ntranscripts", ntranscripts)

        gtfs = []
        for r, pos in intervals.items():

            g = GTF.Entry().copy(gtf)
            g.start, g.end = r
            g.addAttribute("nused", len(pos))
            g.addAttribute("pos", ",".join(["%i:%i" % x for x in pos]))
            gtfs.append(g)

        gtfs.sort(key=lambda x: x.start)

        for g in gtfs:
            options.stdout.write("%s\n" % str(g))

        # check for exon overlap
        intervals = [(g.start, g.end) for g in gtfs]
        nbefore = len(intervals)
        nafter = len(Intervals.combine(intervals))
        if nafter != nbefore:
            noverlapping += 1

        noutput += 1

    if options.loglevel >= 1:
        options.stdlog.write(
            "# ninput=%i, noutput=%i, noverlapping=%i\n" % (ninput, noutput, noverlapping))


def annotatePromoters(iterator, fasta, options):
    """annotate promoters within iterator.

    Entries specied with ``--restrict-source`` are annotated.
    """

    gene_iterator = GTF.gene_iterator(iterator)

    ngenes, ntranscripts, npromotors = 0, 0, 0

    for gene in gene_iterator:
        ngenes += 1
        is_negative_strand = Genomics.IsNegativeStrand(gene[0][0].strand)
        lcontig = fasta.getLength(gene[0][0].contig)
        promotors = []
        transcript_ids = []
        for transcript in gene:

            ntranscripts += 1
            mi, ma = min([x.start for x in transcript]), max(
                [x.end for x in transcript])
            transcript_ids.append(transcript[0].transcript_id)
            # if tss is directly at start/end of contig, the tss will be within an exon.
            # otherwise, it is outside an exon.
            if is_negative_strand:
                promotors.append(
                    (min(lcontig - options.promotor, ma), min(lcontig, ma + options.promotor)))
            else:
                promotors.append(
                    (max(0, mi - options.promotor), max(options.promotor, mi)))

        if options.merge_promotors:
            # merge the promotors (and rename - as sort order might have
            # changed)
            promotors = Intervals.combine(promotors)
            transcript_ids = ["%i" % (x + 1) for x in range(len(promotors))]

        gtf = GTF.Entry()
        gtf.fromGTF(gene[0][0], gene[0][0].gene_id, gene[0][0].gene_id)
        gtf.source = "promotor"

        x = 0
        for start, end in promotors:
            gtf.start, gtf.end = start, end
            gtf.transcript_id = transcript_ids[x]
            options.stdout.write("%s\n" % str(gtf))
            npromotors += 1
            x += 1

    E.info("ngenes=%i, ntranscripts=%i, npromotors=%i" %
           (ngenes, ntranscripts, npromotors))


def annotateRegulons(iterator, fasta, tss, options):
    """annotate regulons within iterator.

    Entries specied with ``--restrict-source`` are annotated.
    """

    gene_iterator = GTF.gene_iterator(iterator)

    ngenes, ntranscripts, nregulons = 0, 0, 0

    upstream, downstream = options.upstream, options.downstream

    for gene in gene_iterator:
        ngenes += 1
        is_negative_strand = Genomics.IsNegativeStrand(gene[0][0].strand)
        lcontig = fasta.getLength(gene[0][0].contig)
        regulons = []
        transcript_ids = []
        for transcript in gene:

            ntranscripts += 1
            mi, ma = min([x.start for x in transcript]), max(
                [x.end for x in transcript])
            if tss:
                # add range to both sides of tss
                if is_negative_strand:
                    interval = ma - options.downstream, ma + options.upstream
                else:
                    interval = mi - options.upstream, mi + options.downstream
            else:
                # add range to both sides of tts
                if is_negative_strand:
                    interval = mi - options.downstream, mi + options.upstream
                else:
                    interval = ma - options.upstream, ma + options.downstream

            interval = (min(lcontig, max(0, interval[0])),
                        min(lcontig, max(0, interval[1])))

            regulons.append(interval)
            transcript_ids.append(transcript[0].transcript_id)

        if options.merge_promotors:
            # merge the regulons (and rename - as sort order might have
            # changed)
            regulons = Intervals.combine(regulons)
            transcript_ids = ["%i" % (x + 1) for x in range(len(regulons))]

        gtf = GTF.Entry()
        gtf.fromGTF(gene[0][0], gene[0][0].gene_id, gene[0][0].gene_id)
        gtf.source = "regulon"

        x = 0
        for start, end in regulons:
            gtf.start, gtf.end = start, end
            gtf.transcript_id = transcript_ids[x]
            options.stdout.write("%s\n" % str(gtf))
            nregulons += 1
            x += 1

    E.info("ngenes=%i, ntranscripts=%i, nregulons=%i" %
           (ngenes, ntranscripts, nregulons))


def annotateGREATDomains(iterator, fasta, options):
    """build great domains

    extend from TSS a basal region.

    """

    gene_iterator = GTF.gene_iterator(iterator)

    counter = E.Counter()

    upstream, downstream = options.upstream, options.downstream
    radius = options.radius
    outfile = options.stdout

    regions = []
    ####################################################################
    # define basal regions for each gene
    # take all basal regions per transcript and merge them
    # Thus, the basal region of a gene might be larger than the sum
    # of options.upstream + options.downstream
    for gene in gene_iterator:
        counter.genes += 1
        is_negative_strand = Genomics.IsNegativeStrand(gene[0][0].strand)

        lcontig = fasta.getLength(gene[0][0].contig)
        regulons = []
        transcript_ids = []

        # collect every basal region per transcript
        for transcript in gene:
            counter.transcripts += 1
            mi, ma = min([x.start for x in transcript]), max(
                [x.end for x in transcript])
            # add range to both sides of tss
            if is_negative_strand:
                interval = ma - options.downstream, ma + options.upstream
            else:
                interval = mi - options.upstream, mi + options.downstream

            interval = (min(lcontig, max(0, interval[0])),
                        min(lcontig, max(0, interval[1])))

            regulons.append(interval)
            transcript_ids.append(transcript[0].transcript_id)

        # take first/last entry
        start, end = min(x[0] for x in regulons), max(x[1] for x in regulons)

        gtf = GTF.Entry()
        gtf.fromGTF(gene[0][0], gene[0][0].gene_id, gene[0][0].gene_id)
        gtf.source = "greatdomain"
        gtf.start, gtf.end = start, end
        regions.append(gtf)

    regions.sort(key=lambda x: (x.contig, x.start))

    outf = IOTools.openFile("test.gff", "w")
    for x in regions:
        outf.write(str(x) + "\n")
    outf.close()

    ####################################################################
    # extend basal regions
    regions.sort(key=lambda x: (x.contig, x.start))

    # iterate within groups of overlapping basal regions
    groups = list(GTF.iterator_overlaps(iter(regions)))
    counter.groups = len(groups)

    last_end = 0
    reset = False

    for region_id, group in enumerate(groups):

        # collect basal intervals in group
        intervals = [(x.start, x.end) for x in group]

        def overlapsBasalRegion(pos):
            for start, end in intervals:
                if start == pos or end == pos:
                    continue
                if start <= pos < end:
                    return True
                if start > pos:
                    return False
            return False

        # deal with boundary cases - end of contig
        if region_id < len(groups) - 1:
            nxt = groups[region_id + 1]
            if nxt[0].contig == group[0].contig:
                next_start = min([x.start for x in nxt])
            else:
                next_start = fasta.getLength(group[0].contig)
                reset = True
        else:
            next_start = fasta.getLength(group[0].contig)
            reset = True

        # last_end = basal extension of previous group
        # next_start = basal_extension of next group

        # extend region to previous/next group always extend
        # dowstream, but upstream only extend if basal region of an
        # interval is not already overlapping another basal region
        # within the group
        save_end = 0
        for gtf in group:
            save_end = max(save_end, gtf.end)
            if gtf.strand == "+":
                if not overlapsBasalRegion(gtf.start):
                    gtf.start = max(gtf.start - radius, last_end)
                # always extend downstream
                gtf.end = min(gtf.end + radius, next_start)
            else:
                # always extend downstream
                gtf.start = max(gtf.start - radius, last_end)
                if not overlapsBasalRegion(gtf.end):
                    gtf.end = min(gtf.end + radius, next_start)
            outfile.write(str(gtf) + "\n")
            counter.regulons += 1

        if len(group) > 1:
            counter.overlaps += len(group)
        else:
            counter.nonoverlaps += 1

        if reset:
            last_end = 0
            reset = False
        else:
            last_end = save_end

    E.info("%s" % str(counter))


def annotateTTS(iterator, fasta, options):
    """annotate termination sites within iterator.

    Entries specified with ``--restrict-source are annotated``.
    """

    gene_iterator = GTF.gene_iterator(iterator)

    ngenes, ntranscripts, npromotors = 0, 0, 0

    for gene in gene_iterator:
        ngenes += 1
        is_negative_strand = Genomics.IsNegativeStrand(gene[0][0].strand)
        lcontig = fasta.getLength(gene[0][0].contig)
        tts = []
        transcript_ids = []
        for transcript in gene:

            ntranscripts += 1
            mi, ma = min([x.start for x in transcript]), max(
                [x.end for x in transcript])
            transcript_ids.append(transcript[0].transcript_id)
            # if tts is directly at start/end of contig, the tss will
            # be within an exon.  otherwise, it is outside an exon.
            if is_negative_strand:
                tts.append(
                    (max(0, mi - options.promotor), max(options.promotor, mi)))
            else:
                tts.append(
                    (min(ma, lcontig - options.promotor),
                     min(lcontig, ma + options.promotor)))

        if options.merge_promotors:
            # merge the promotors (and rename - as sort order might have
            # changed)
            tts = Intervals.combine(tts)
            transcript_ids = ["%i" % (x + 1) for x in range(len(tts))]

        gtf = GTF.Entry()
        gtf.fromGTF(gene[0][0], gene[0][0].gene_id, gene[0][0].gene_id)
        gtf.source = "tts"

        x = 0
        for start, end in tts:
            gtf.start, gtf.end = start, end
            gtf.transcript_id = transcript_ids[x]
            options.stdout.write("%s\n" % str(gtf))
            npromotors += 1
            x += 1

    if options.loglevel >= 1:
        options.stdlog.write(
            "# ngenes=%i, ntranscripts=%i, ntss=%i\n" %
            (ngenes, ntranscripts, npromotors))


def annotateGenes(iterator, fasta, options):
    """annotate gene structures

    This method outputs intervals for first/middle/last exon/intron,
    UTRs and flanking regions.

    This method annotates per transcript. In order to achieve a unique tiling,
    use only a single transcript per gene and remove any overlap between
    genes.

    """

    gene_iterator = GTF.gene_iterator(iterator)

    ngenes, ntranscripts, nskipped = 0, 0, 0

    results = []
    increment = options.increment

    introns_detail = "introns" in options.detail
    exons_detail = "exons" in options.detail

    for gene in gene_iterator:
        ngenes += 1
        is_negative_strand = Genomics.IsNegativeStrand(gene[0][0].strand)
        try:
            lcontig = fasta.getLength(gene[0][0].contig)
        except KeyError:
            nskipped += 1
            continue

        results = []

        for transcript in gene:

            def _add(interval, anno):
                gtf = GTF.Entry()
                gtf.contig = transcript[0].contig
                gtf.gene_id = transcript[0].gene_id
                gtf.transcript_id = transcript[0].transcript_id
                gtf.strand = transcript[0].strand
                gtf.feature = anno
                gtf.start, gtf.end = interval
                results.append(gtf)

            ntranscripts += 1

            exons = [(x.start, x.end)
                     for x in transcript if x.feature == "exon"]
            if len(exons) == 0:
                nskipped += 1

            exons.sort()
            introns = []
            end = exons[0][1]
            for exon in exons[1:]:
                introns.append((end, exon[0]))
                end = exon[1]

            # add flank
            start, end = exons[0][0], exons[-1][1]
            upstream, downstream = [], []
            for x in range(0, options.flank, increment):
                upstream.append((start - increment, start))
                start -= increment
                downstream.append((end, end + increment))
                end += increment

            # remove out-of-bounds coordinates
            upstream = [x for x in upstream if x[0] >= 0]
            downstream = [x for x in downstream if x[1] <= lcontig]

            if is_negative_strand:
                exons.reverse()
                introns.reverse()
                upstream, downstream = downstream, upstream

            # add exons
            if exons_detail:
                _add(exons[0], "first_exon")
                if len(exons) > 1:
                    _add(exons[-1], "last_exon")
                for e in exons[1:-1]:
                    _add(e, "middle_exon")
            else:
                for e in exons:
                    _add(e, "exon")

            # add introns
            if introns_detail:
                if len(introns) > 0:
                    _add(introns[0], "first_intron")
                if len(introns) > 1:
                    _add(introns[-1], "last_intron")
                for i in introns[1:-1]:
                    _add(i, "middle_intron")
            else:
                for i in introns:
                    _add(i, "intron")

            for x, u in enumerate(upstream):
                _add(u, "upstream_%i" % (increment * (x + 1)))

            for x, u in enumerate(downstream):
                _add(u, "downstream_%i" % (increment * (x + 1)))

            results.sort(key=lambda x: x.feature)

        cache = []
        for key, vals in itertools.groupby(results, key=lambda x: x.feature):
            v = list(vals)
            intervals = [(x.start, x.end) for x in v]
            intervals = Intervals.combine(intervals)

            for start, end in intervals:
                r = GTF.Entry()
                r.copy(v[0])
                r.start, r.end = start, end
                cache.append(r)

        cache.sort(key=lambda x: x.start)
        for r in cache:
            options.stdout.write("%s\n" % str(r))

    E.info("ngenes=%i, ntranscripts=%i, nskipped=%i\n" %
           (ngenes, ntranscripts, nskipped))


def main(argv=None):

    if not argv:
        argv = sys.argv

    parser = E.OptionParser(
        version="%prog version: $Id$",
        usage=globals()["__doc__"])

    parser.add_option("-g", "--genome-file", dest="genome_file", type="string",
                      help="filename with genome [default=%default].")

    parser.add_option("-i", "--ignore-missing", dest="ignore_missing",
                      action="store_true",
                      help="Ignore transcripts on contigs that are not "
                      "in the genome-file [default=%default].")

    parser.add_option("-s", "--restrict-source", dest="restrict_source",
                      type="choice",
                      choices=("protein_coding", "pseudogene", "lncRNA"),
                      help="restrict input by source [default=%default].")

    parser.add_option("-m", "--method", dest="method", type="choice",
                      choices=("full", "genome", "exons",
                               "promotors", "tts",
                               "regulons", "tts-regulons",
                               "genes",
                               "territories", "tss-territories",
                               "great-domains",
                               ),
                      help="method for defining segments [default=%default].")

    parser.add_option(
        "-r", "--territory-extension", dest="radius", type="int",
        help="radius of a territory [default=%default].")

    parser.add_option(
        "-f", "--flank-size", dest="flank", type="int",
        help="size of the flanking region next to a gene [default=%default].")

    parser.add_option(
        "--flank-increment-size", dest="increment", type="int",
        help="size of increment in flank in genestructure annotation "
        "[default=%default].")

    parser.add_option(
        "-p", "--promotor-size", dest="promotor", type="int",
        help="size of a promotor region [default=%default].")

    parser.add_option(
        "-u", "--upstream-extension", dest="upstream", type="int",
        help="size of region upstream of tss [default=%default].")

    parser.add_option(
        "-d", "--downstream-extension", dest="downstream", type="int",
        help="size of region downstream of tss [default=%default].")

    parser.add_option(
        "--gene-detail", dest="detail", type="choice",
        choices=("introns+exons", "exons", "introns"),
        help="level of detail for gene structure annotation "
        "[default=%default].")

    parser.add_option(
        "--merge-overlapping-promotors", dest="merge_promotors",
        action="store_true",
        help="merge overlapping promotors [default=%default].")

    parser.add_option(
        "--min-intron-length", dest="min_intron_length",
        type="int",
        help="minimum intron length. If the distance between two "
        "consecutive exons is smaller, the region will be marked "
        "'unknown' [default=%default].")

    parser.add_option(
        "--is-unsorted", dest="is_sorted", action="store_false",
        help="sort input before processing. Otherwise, the input is assumed "
        "to be sorted [default=%default].")

    parser.set_defaults(
        genome_file=None,
        flank=1000,
        increment=1000,
        max_frameshift_length=4,
        min_intron_length=30,
        ignore_missing=False,
        restrict_source=None,
        method="genome",
        radius=50000,
        promotor=5000,
        merge_promotors=False,
        upstream=5000,
        downstream=5000,
        detail="exons",
        is_sorted=True,
    )

    (options, args) = E.Start(parser)

    if options.genome_file:
        fasta = IndexedFasta.IndexedFasta(options.genome_file)
    else:
        raise ValueError("please specify a --genome-file")

    if not options.restrict_source:
        iterator = GTF.iterator(options.stdin)

    elif options.restrict_source:
        iterator = GTF.iterator_filtered(GTF.iterator(options.stdin),
                                         source=options.restrict_source)

    # elif options.method in ("promotors", "tts", "regulons"):
    #     iterator = GTF.iterator_filtered( GTF.iterator(options.stdin), source = "protein_coding")
    # else:
    #     iterator = GTF.iterator(options.stdin)

    if not options.is_sorted:
        iterator = GTF.iterator_sorted(iterator, sort_order="position")

    if options.method == "full" or options.method == "genome":
        segmentor = annotateGenome(iterator, fasta, options)
    elif options.method == "territories":
        segmentor = buildTerritories(iterator, fasta, 'gene', options)
    elif options.method == "tss-territories":
        segmentor = buildTerritories(iterator, fasta, 'tss', options)
    elif options.method == "exons":
        segmentor = annotateExons(iterator, fasta, options)
    elif options.method == "promotors":
        segmentor = annotatePromoters(iterator, fasta, options)
    elif options.method == "regulons":
        segmentor = annotateRegulons(iterator, fasta, True, options)
    elif options.method == "tts-regulons":
        segmentor = annotateRegulons(iterator, fasta, False, options)
    elif options.method == "tts":
        segmentor = annotateTTS(iterator, fasta, options)
    elif options.method == "genes":
        segmentor = annotateGenes(iterator, fasta, options)
    elif options.method == "great-domains":
        segmentor = annotateGREATDomains(iterator, fasta, options)

    E.Stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



cdef extern from "string.h":
  ctypedef int size_t
  void *memcpy(void *dst,void *src,size_t len)
  void *memmove(void *dst,void *src,size_t len)
  void *memset(void *b,int c,size_t len)

cdef extern from "stdlib.h":
  void free(void *)
  void *malloc(size_t)
  void *calloc(size_t,size_t)
  void *realloc(void *,size_t)
  int c_abs "abs" (int)
  void qsort(void *base, size_t nmemb, size_t size,
             int (*compar)(void *,void *))

cdef extern from "stdio.h":
  ctypedef struct FILE:
    pass
  FILE *fopen(char *,char *)
  int fclose(FILE *)
  int sscanf(char *str,char *fmt,...)
  int sprintf(char *str,char *fmt,...)
  int fprintf(FILE *ifile,char *fmt,...)
  char *fgets(char *str,int size,FILE *ifile)

cdef extern from "string.h":
  int strcmp(char *s1, char *s2)
  int strncmp(char *s1,char *s2,size_t len)
  char *strcpy(char *dest,char *src)
  char *strdup(char *)
  char *strcat(char *,char *)

cdef extern from "intervaldb.h":
  ctypedef struct IntervalMap:
    int start
    int end
    int target_id
    int sublist

  ctypedef struct IntervalIndex:
    int start
    int end

  ctypedef struct SublistHeader:
    int start
    int len

  ctypedef struct SubheaderFile:
    pass
  
  ctypedef struct IntervalDBFile:
    int n
    int ntop
    int nlists
    int div
    int nii
    IntervalIndex *ii
    SublistHeader *subheader
    SubheaderFile subheader_file
    FILE *ifile_idb

  ctypedef struct IntervalIterator:
    pass

  ctypedef struct FilePtrRecord:
    FILE *ifile
    int left
    int right
    int ihead
    char *filename

  int imstart_qsort_cmp(void *void_a,void *void_b)
  int target_qsort_cmp(void *void_a,void *void_b)
  IntervalMap *read_intervals(int n,FILE *ifile) except NULL
  SublistHeader *build_nested_list(IntervalMap im[],int n,int *p_n,int *p_nlists) except NULL
  SublistHeader *build_nested_list_inplace(IntervalMap im[],int n,int *p_n,int *p_nlists) except NULL
  IntervalMap *interval_map_alloc(int n) except NULL
  IntervalIterator *interval_iterator_alloc() except NULL
  int free_interval_iterator(IntervalIterator *it)
  IntervalIterator *reset_interval_iterator(IntervalIterator *it)
  int find_intervals(IntervalIterator *it0,int start,int end,IntervalMap im[],int n,SublistHeader subheader[],int nlists,IntervalMap buf[],int nbuf,int *p_nreturn,IntervalIterator **it_return) except -1
  char *write_binary_files(IntervalMap im[],int n,int ntop,int div,SublistHeader *subheader,int nlists,char filestem[])
  IntervalDBFile *read_binary_files(char filestem[],char err_msg[],int subheader_nblock) except NULL
  int free_interval_dbfile(IntervalDBFile *db_file)
  int find_file_intervals(IntervalIterator *it0,int start,int end,IntervalIndex ii[],int nii,SublistHeader subheader[],int nlists,SubheaderFile *subheader_file,int ntop,int div,FILE *ifile,IntervalMap buf[],int nbuf,int *p_nreturn,IntervalIterator **it_return) except -1
  int write_padded_binary(IntervalMap im[],int n,int div,FILE *ifile)
  int read_imdiv(FILE *ifile,IntervalMap imdiv[],int div,int i_div,int ntop)
  int save_text_file(char filestem[],char basestem[],char err_msg[],FILE *ofile)
  int text_file_to_binaries(FILE *infile,char err_msg[])
  int C_int_max

cdef class IntervalDB:
  cdef int n
  cdef int ntop
  cdef int nlists
  cdef IntervalMap *im
  cdef SublistHeader *subheader

cdef class IntervalDBIterator:
  cdef IntervalIterator *it,*it_alloc
  cdef IntervalMap im_buf[1024]
  cdef int ihit,nhit,start,end
  cdef IntervalDB db

  cdef int cnext(self)

cdef class IntervalFileDB:
  cdef IntervalDBFile *db

cdef class IntervalFileDBIterator:
  cdef IntervalIterator *it,*it_alloc
  cdef IntervalMap *im_buf
  cdef int ihit,nhit,start,end,nbuf
  cdef IntervalFileDB db
  cdef IntervalDB idb

  cdef int restart(self,int start,int end,IntervalFileDB db) except -2
  cdef int reset(self) except -2
  cdef int cnext(self,int *pkeep)
  cdef int extend(self,int ikeep)
  cdef int saveInterval(self,int start,int end,int target_id)
  cdef int nextBlock(self,int *pkeep) except -2
  cdef IntervalMap *getIntervalMap(self)
  cdef int loadAll(self) except -1
  cdef int copy(self,IntervalFileDBIterator src)






"""
Histogram2D.py - functions for handling two-dimensional histograms.
===================================================================

:Tags: Python

"""

import string


def Calculate(values, mode=0, bin_function=None):
    """Return a list of (value, count) pairs, summarizing the input values.
    Sorted by increasing value, or if mode=1, by decreasing count.

    If bin_function is given, map it over values first.

    """

    if bin_function:
        values = list(map(bin_function, values))

    bins = {}
    for val in values:
        v = "%f-%f" % tuple(val)
        bins[v] = bins.get(v, 0) + 1

    bb = list(bins.items())

    if mode:
        bb.sort(lambda x, y: cmp(y[1], x[1]))
    else:
        bb.sort()

    r = []
    for v, n in bb:
        x, y = list(map(string.atof, string.split(v, "-")))
        r.append((x, y, n))

    return r


def Print(h, bin_function=None):
    """print a histogram.

    A histogram can either be a list/tuple of values or
    a list/tuple of lists/tuples where the first value contains
    the bin and second contains the values (which can again be
    a list/tuple).

    :param format: output format. 
        0 = print histogram in several lines,
        1 = print histogram on single line
    """

    if bin_function:
        h = list(map(bin_function, h))

    for hh in h:
        print(string.join(list(map(str, hh)), "\t"))



'''
WrapperZinba.py - utility functions for zinba output
====================================================

:Tags: Python

Purpose
-------

Usage
-----

Documentation
-------------

Code
----

'''

import collections

ZinbaPeak = collections.namedtuple(
    "ZinbaPeak",
    "contig unrefined_start unrefined_end strand posterior "
    "summit height refined_start refined_end median fdr")


def iteratePeaks(infile):
    '''iterate of zinba peaks in infile.'''

    for line in infile:

        if line.startswith("#"):
            continue
        if line.startswith("PEAKID\tChrom"):
            continue
        # skip empty lines
        if line.startswith("\n"):
            continue

        data = line[:-1].split("\t")

        if len(data) != 12:
            raise ValueError("could not parse line %s" % line)

        # I assume these are 1-based coordinates
        data[2] = max(int(data[2]) - 1, 0)
        # end
        data[3] = int(data[3])
        # posterior
        data[5] = float(data[5])
        # summit
        data[6] = max(int(data[6]) - 1, 0)
        # height
        data[7] = int(data[7])
        # refined_start
        data[8] = max(int(data[8]) - 1, 0)
        # end
        data[9] = int(data[9])
        # median
        data[10] = int(data[10])
        # qvalue
        data[11] = float(data[11])

        yield ZinbaPeak._make(data[1:])



"""
Histogram.py - Various functions to deal with histograms
===========================================================

:Author:
:Tags: Python

Histograms can be calculated from a list/tuple/array of
values. The histogram returned is then a list of tuples
of the format [(bin1,value1), (bin2,value2), ...].

"""

import sys
import re
import math
import scipy
import scipy.stats
import bisect
import numpy
from functools import reduce


def CalculateFromTable(dbhandle,
                       field_name,
                       from_statement,
                       num_bins=None,
                       min_value=None,
                       max_value=None,
                       intervals=None,
                       increment=None):
    """get a histogram using an SQL-statement.
    Intervals can be either supplied directly or are build
    from the data by providing the number of bins and optionally
    a minimum or maximum value.

    If no number of bins are provided, the bin-size is 1.

    This command uses the INTERVAL command from MYSQL, i.e. a bin value
    determines the upper boundary of a bin.
    """

    if not min_value:
        min_value = int(math.floor(dbhandle.Execute(
            "SELECT MIN(%s) %s" % (field_name, from_statement)).fetchone()[0]))

    if not max_value:
        max_value = int(math.ceil(dbhandle.Execute(
            "SELECT MAX(%s) %s" % (field_name, from_statement)).fetchone()[0]))

    if increment:
        step_size = increment
    elif num_bins:
        step_size = int(float(max_value - min_value) / float(num_bins))
    else:
        step_size = 1

    if not intervals:
        intervals = list(range(min_value, max_value, step_size))

    i_string = ",".join(list(map(str, intervals)))

    statement = "SELECT INTERVAL( %s, %s )-1 AS i, COUNT(*) %s GROUP BY i" % (
        field_name, i_string, from_statement)

    return Convert(dbhandle.Execute(statement).fetchall(), intervals)


def CalculateConst(values,
                   num_bins=None,
                   min_value=None,
                   max_value=None,
                   intervals=None,
                   increment=None,
                   combine=None):
    """calculate a histogram based on a list or tuple of values.
    """

    if not min_value:
        min_value = int(math.floor(min(values)))

    if not max_value:
        max_value = int(math.ceil(max(values))) + 1

    if increment:
        step_size = increment
    elif num_bins:
        step_size = int(float(max_value - min_value) / float(num_bins))
    else:
        step_size = 1

    if not intervals:
        intervals = list(range(min_value, max_value, step_size))

    histogram = [0] * len(intervals)
    for v in values:
        i = 0
        while i < len(intervals) and v > intervals[i]:
            i += 1
        if i < len(intervals):
            histogram[i] += 1

    return intervals, histogram


def Calculate(values,
              num_bins=None,
              min_value=None,
              max_value=None,
              intervals=None,
              increment=None,
              combine=None,
              no_empty_bins=0,
              dynamic_bins=False,
              ignore_out_of_range=True):
    """calculate a histogram based on a list or tuple of values.

    use scipy for calculation.
    """

    if len(values) == 0:
        return []

    if not intervals:

        if min_value is None:
            min_value = min(values)

        if max_value is None:
            max_value = max(values)

        if dynamic_bins:
            intervals = list(
                set([x for x in values if min_value <= x <= max_value]))
            intervals.sort()
        else:
            if increment:
                step_size = increment
            elif num_bins and max_value:
                step_size = float(max_value - min_value) / float(num_bins)
            else:
                step_size = 1.0

            num_bins = int(
                math.ceil((float(max_value) - float(min_value)) / float(step_size)))
            intervals = [float(min_value) + float(x) * float(step_size)
                         for x in range(num_bins + 1)]

    if not ignore_out_of_range:
        new_values = []
        for v in values:
            if v < min_value:
                v = min_value
            elif v > max_value:
                v = max_value
            new_values.append(v)

        values = new_values

    return Convert(scipy.stats.histogram2(values, intervals), intervals, no_empty_bins)


def Scale(h, scale=1.0):
    """rescale bins in histogram.
    """
    n = []
    for b, v in h:
        n.append((b * scale, v))
    return n


def Convert(h, i, no_empty_bins=0):
    """add bins to histogram.
    """
    n = []
    for x in range(0, len(h)):
        if no_empty_bins and h[x] == 0:
            continue
        n.append((i[x], h[x]))
    return n


def Combine(source_histograms, missing_value=0):
    """combine a list of histograms
    Each histogram is a sorted list of bins and counts.
    The counts can be tuples.
    """

    new_bins = {}

    # get all bins
    for h in source_histograms:
        for data in h:
            new_bins[data[0]] = []

    # add values
    length = 1
    l = 0
    for h in source_histograms:

        # add data for used bins
        for data in h:
            bin = data[0]
            if len(data) != 2:
                v = data[1:]
            else:
                v = data[1]

            if isinstance(v, list) or isinstance(v, tuple):
                l = len(v)
                for x in v:
                    new_bins[bin].append(x)
            else:
                l = 1
                new_bins[bin].append(v)

        # add missing value for unused bins
        for b in list(new_bins.keys()):
            if len(new_bins[b]) < length:
                for x in range(0, l):
                    new_bins[b].append(missing_value)

        length += l

    return __ConvertToList(new_bins)


def Print(h, intervalls=None, format=0, nonull=None, format_value=None, format_bin=None):
    """print a histogram.

    A histogram can either be a list/tuple of values or
    a list/tuple of lists/tuples where the first value contains
    the bin and second contains the values (which can again be
    a list/tuple).

    format
       0 = print histogram in several lines
       1 = print histogram on single line

    """

    Write(sys.stdout, h, intervalls, format, nonull, format_value, format_bin)


def Write(outfile, h, intervalls=None, format=0, nonull=None,
          format_value=None, format_bin=None):
    """print a histogram.

    A histogram can either be a list/tuple of values or
    a list/tuple of lists/tuples where the first value contains
    the bin and second contains the values (which can again be
    a list/tuple).

    :param format: output format.
        0 = print histogram in several lines,
        1 = print histogram on single line

    """

    lines = []

    if len(h) == 0:
        return

    if format_value:
        def fv(x):
            if x == "na":
                return x
            else:
                return format_value % x
    else:
        fv = str

    if format_bin:
        fb = lambda x: format_bin % x
    else:
        fb = str

    if not isinstance(h[0], list) and not isinstance(h[0], tuple):
        for x in range(0, len(h)):

            if intervalls:
                lines.append(fb(intervalls[x]) + "\t" + fv(h[x]))
            else:
                lines.append(fb(x) + "\t" + fv(h[x]))
    else:
        for x, v in h:
            if isinstance(v, list) or isinstance(v, tuple):
                val = "\t".join(list(map(fv, v)))
            else:
                val = fv(v)

            if intervalls:
                lines.append(fb(intervalls[x]) + "\t" + val)
            else:
                lines.append(fb(x) + "\t" + val)

    # print values
    if nonull is not None:
        for l in range(0, len(lines)):
            lines[l] = re.sub("\t0", "\t%s" % nonull, lines[l])

    if format == 0:
        outfile.write("\n".join(lines) + "\n")
    elif format == 1:
        outfile.write(" | ".join(lines) + "\n")
    else:
        outfile.write("\n".join(lines) + "\n")


def Fill(h):
    """fill every empty value in histogram with
    previous value.
    """

    new_h = []

    x, v = h[0]

    if isinstance(v, list) or isinstance(v, tuple):
        l = len(v)
        previous_v = [0] * l
    else:
        previous_v = 0

    for x, v in h:
        if isinstance(v, list) or isinstance(v, tuple):
            for i in range(0, l):
                if v[i] == 0:
                    v[i] = previous_v[i]
                else:
                    previous_v[i] = v[i]
        else:
            if v == 0:
                v = previous_v
            else:
                previous_v = v

        new_h.append((x, v))

    return new_h


def Normalize(h):

    # first count totals
    if isinstance(h[0][1], list) or isinstance(h[0][1], tuple):
        l = len(h[0][1])
        is_list = 1
    else:
        is_list = 0
        l = 1

    totals = [0.0] * l

    for bin, v in h:
        if is_list:
            for x in range(0, l):
                try:
                    totals[x] += v[x]
                except TypeError:
                    pass
        else:
            totals[0] += v

    for x in range(0, l):
        if totals[x] == 0:
            totals[x] = 1

    # first count totals
    new_histogram = []
    for bin, v in h:
        if is_list:
            vv = []
            for x in range(0, l):
                vv.append(float(v[x]) / totals[x])
        else:
            vv = float(v) / totals[0]

        new_histogram.append((bin, vv))

    return new_histogram


def Add(h1, h2):
    """adds values of histogram h1 and h2 and
    returns a new histogram
    """

    new_bins = {}
    # get all bins
    for h in (h1, h2):
        if h:
            for bin, v in h:
                new_bins[bin] = 0

    for h in (h1, h2):
        if h:
            for bin, v in h:
                new_bins[bin] += v

    return __ConvertToList(new_bins)


def __ConvertToList(new_bins):
    """converts a hash to a histogram.
    """

    # convert to list
    keys = list(new_bins.keys())
    keys.sort()
    new_histogram = []

    for k in keys:
        new_histogram.append((k, new_bins[k]))

    return new_histogram


def SmoothWrap(histogram, window_size):
    """smooth histogram by sliding window-method, where
    the window is wrapped around the borders. The sum of
    all values is entered at center of window.
    """

    new_histogram = [0] * len(histogram)

    half_window_size = window_size / 2
    length = len(histogram)

    # 1. start with window
    cumul = 0
    for i in range(length - half_window_size, length):
        cumul = cumul + histogram[i]
    for i in range(0, half_window_size + 1):
        cumul = cumul + histogram[i]

    # 2. iterate over histogram and add values over windows_size
    y = length - half_window_size
    z = half_window_size
    for i in range(0, length):

        new_histogram[i] = cumul

        y = y + 1
        z = z + 1
        if y >= length:
            y = 0
        if z >= length:
            z = 0
        cumul = cumul - histogram[y] + histogram[z]

    return new_histogram


def GetMaximumIndex(histogram):
    return histogram.index(max(histogram))


def GetMinimumIndex(histogram):
    return histogram.index(min(histogram))


def PrintAscii(histogram, step_size=1):
    """print histogram ascii-style.
    """

    l = len(histogram)
    m = max(histogram)

    if m == 0:
        print("----> histogram is empty")
        return

    f = 100.0 / m

    print("----> histogram: len=%i, max=%i" % (l, m))
    for x in range(1, l, step_size):
        s = "|"
        s += " " * (int(histogram[x] * f) - 1) + "*"

        print("%5i" % x, s)


def Count(data):
    """count categorized data. Returns a list
    of tuples with (count, token).
    """

    counts = []
    data.sort()

    last_c = None
    for c in data:
        if last_c != c:
            if last_c:
                counts.append((t, last_c))
            last_c = c
            t = 0
        t += 1

    counts.append((t, last_c))

    return counts


def Accumulate(h, num_bins=2, direction=1):
    """add successive counts in histogram.
    Bins are labelled by group average.
    """

    if len(h) == []:
        return []

    new_histogram = []

    if isinstance(h[0][1], list) or isinstance(h[0][1], tuple):
        l = len(h[0][1])
        is_list = 1
    else:
        is_list = 0
        l = 1

    if direction != 1:
        h.reverse()

    i = 0
    if is_list:
        vv = [0] * l
        bb = 0
        for b, v in h:

            bb += b
            for x in range(0, l):
                vv[x] += v[x]

            i += 1
            if i % num_bins == 0:
                new_histogram.append((float(bb) / float(num_bins), vv))
                vv = [0] * l
                bb = 0

        if (i % num_bins):
            new_histogram.append((float(bb) / float(i % num_bins), vv))
    else:
        vv = 0
        bb = 0
        for b, v in h:
            bb += b
            vv += v

            i += 1
            if i % num_bins == 0:
                new_histogram.append(float(bb) / float(num_bins), vv)
                vv = 0
                bb = 0

        if (i % num_bins):
            new_histogram.append(float(bb) / float(i % num_bins), vv)

    # reorder h and new_histogram
    if direction != 1:
        h.reverse()
        new_histogram.reverse()

    return new_histogram


def Cumulate(h, direction=1):
    """calculate cumulative distribution.
    """

    if len(h) == []:
        return []

    new_histogram = []

    if isinstance(h[0][1], list) or isinstance(h[0][1], tuple):
        l = len(h[0][1])
        is_list = 1
    else:
        is_list = 0
        l = 1

    if direction != 1:
        h.reverse()

    if is_list:
        vv = [0] * l
        for b, v in h:
            for x in range(0, l):
                vv[x] += v[x]
            new_histogram.append((b, [x for x in vv]))
    else:
        vv = 0
        for b, v in h:
            vv += v
            new_histogram.append((b, vv))

    # reorder h and new_histogram
    if direction != 1:
        h.reverse()
        new_histogram.reverse()

    return new_histogram


def AddRelativeAndCumulativeDistributions(h):
    """adds relative and cumulative percents to a histogram.
    """

    if len(h) == []:
        return []

    new_histogram = []
    total = float(reduce(lambda x, y: x + y, [x[1] for x in h]))

    cumul_down = int(total)
    cumul_up = 0

    for bin, val in h:
        percent = float(val) / total
        cumul_up += val
        percent_cumul_up = float(cumul_up) / total
        percent_cumul_down = float(cumul_down) / total

        new_histogram.append(
            (bin, (val, percent, cumul_up, percent_cumul_up, cumul_down, percent_cumul_down)))
        cumul_down -= val

    return new_histogram


def histogram(values, mode=0, bin_function=None):
    """Return a list of (value, count) pairs, summarizing the input values.
    Sorted by increasing value, or if mode=1, by decreasing count.
    If bin_function is given, map it over values first.
    Ex: vals = [100, 110, 160, 200, 160, 110, 200, 200, 220]
    histogram(vals) ==> [(100, 1), (110, 2), (160, 2), (200, 3), (220, 1)]
    histogram(vals, 1) ==> [(200, 3), (160, 2), (110, 2), (100, 1), (220, 1)]
    histogram(vals, 1, lambda v: round(v, -2)) ==> [(200.0, 6), (100.0, 3)]"""

    if bin_function:
        values = list(map(bin_function, values))
    bins = {}
    for val in values:
        bins[val] = bins.get(val, 0) + 1
        if mode:
            return sort(list(bins.items()), lambda x, y: cmp(y[1], x[1]))
        else:
            return sort(list(bins.items()))


def cumulate(histogram):
    """cumulate histogram in place.

    histogram is list of (bin, value) or (bin, (values,) )
    """

    if len(histogram) == 0:
        return

    if isinstance(h[0][1], list) or isinstance(h[0][1], tuple):
        n = len(histogram[0][1])
        l = [0] * n
        for bin, vv in histogram:
            for x in range(n):
                vv[x] += l[x]
                l[x] = vv[x]
    else:
        l = 0
        for x in range(len(histogram)):
            histogram[x] = (histogram[x][0], histogram[x][1] + l)
            l = histogram[x][1]


def normalize(histogram):
    """normalize histogram in place.

    histogram is list of (bin, value) or (bin, (values,) )
    """

    if len(histogram) == 0:
        return

    if isinstance(histogram[0][1], list) or isinstance(histogram[0][1], tuple):
        n = len(histogram[0][1])
        m = [0] * n
        for bin, d in histogram:
            for x in range(n):
                m[x] = max(m[x], d[x])
        for bin, d in histogram:
            for x in range(n):
                if m[x] > 0:
                    d[x] = float(d[x]) / m[x]
    else:
        m = float(max([x[1] for x in histogram]))
        if m > 0:
            for x in range(len(histogram)):
                histogram[x] = (histogram[x][0], histogram[x][1] / m)


def fill(iterator, bins):
    """fill a histogram from bins. 

    The values are given by an iterator so that the histogram
    can be built on the fly.

    Description:

    Count the number of times values from array a fall into
    numerical ranges defined by bins.  Range x is given by
    bins[x] <= range_x < bins[x+1] where x =0,N and N is the
    length of the bins array.  The last range is given by
    bins[N] <= range_N < infinity.  Values less than bins[0] are
    not included in the histogram.

    Arguments:
       iterator -- The iterator.
       bins -- 1D array.  Defines the ranges of values to use during
       histogramming.

    Returns:
    1D array.  Each value represents the occurences for a given
    bin (range) of values.
    """

    h = numpy.zeros(len(bins))

    # bins might not be uniform, so do a bisect
    for value in iterator:
        i = bisect.bisect_left(value)
        h[i] += 1

    return h


def fillHistograms(infile, columns, bins):
    """fill several histograms from several columns in a file.

    The histograms are built on the fly.

    Description:

    Count the number of times values from array a fall into
    numerical ranges defined by bins.  Range x is given by
    bins[x] <= range_x < bins[x+1] where x =0,N and N is the
    length of the bins array.  The last range is given by
    bins[N] <= range_N < infinity.  Values less than bins[0] are
    not included in the histogram.

    Arguments:
       file -- The input file.
       columns -- columns to use
       bins -- a list of 1D arrays.  Defines the ranges of values to use during
       histogramming.

    Returns:
    a list of 1D arrays.  Each value represents the occurences for a given
    bin (range) of values.

    WARNING: missing value in columns are ignored
    """

    assert(len(bins) == len(columns))

    hh = [numpy.zeros(len(bins[x]), numpy.float) for x in range(len(columns))]

    for line in infile:
        if line[0] == "#":
            continue
        data = line[:-1].split()
        # bins might not be uniform, so do a bisect
        for x, y in enumerate(columns):
            try:
                v = float(data[y])
            except IndexError:
                continue
            i = bisect.bisect(bins[x], v) - 1
            if i >= 0:
                hh[x][i] += 1

    return hh



'''beds2counts - compute overlap stats between multiple bed files
=================================================================

:Tags: Genomics Intervals Comparison BED Counting

Purpose
-------

This script takes multiple bed files e.g. from multiple samples from
the same experiment. It assesses the overlap between samples and
outputs a count for each merged interval corresponding to the number
of samples that a particular interval was found in.


Example
-------

For example if the command::

    cgat bed2counts a.bed b.bed c.bed > output.tsv

is run, where a.bed-c.bed look like::

                     1         2         3         4
           012345678901234567890123456789012345678901234
    a.bed: -------          -----               -------
    b.bed:      -----        --
    c.bed:  ---

    Union: ----------       -----               -------

Then output.tsv will look like::

    contig	start	end	count
    chr1	0	7	3
    chr1	17	22	2
    chr1	37	44	1

Options
-------

The only option other than the standard cgat options is -i, --bed-file this
allows the input files to be provided as a comma seperated list to the option
rather than a space delimited set of positional arguements. It is present
purely for galaxy compatibility.

Usage
-----

    cgat beds2counts BED [BED ...] [OPTIONS]

Command line options
--------------------

'''
import tempfile
import sys

try:
    import pybedtools
except ImportError:
    pass

import CGAT.Experiment as E
import CGAT.Bed as Bed
import collections
import CGAT.IOTools as IOTools
import CGAT.IndexedGenome as IndexedGenome


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv:
        argv = sys.argv

    # setup command line parser
    parser = E.OptionParser(version="%prog version: $Id",
                            usage=globals()["__doc__"])

    parser.add_option(
        "--bed-file", dest="infiles", type="string",
        metavar="bed",
        help="supply list of bed files",
        action="append")

    parser.set_defaults(infiles=[])

    # add common options (-h/--help, ...) and parse command line
    (options, args) = E.Start(parser, argv=argv)

    options.infiles.extend(args)
    if len(options.infiles) == 0:
        raise ValueError('please provide at least 1 bed file')

    E.info("concatenating bed files")
    # concatenate the list of files
    tmp = tempfile.NamedTemporaryFile(delete=False, mode="w")
    tmp_merge = tempfile.NamedTemporaryFile(delete=False, mode="w")
    infs = options.infiles
    for inf in infs:
        for bed in Bed.iterator(IOTools.openFile(inf)):
            tmp.write("%s\n" % bed)
    tmp.close()

    E.info("merging bed entries")
    # merge the bed entries in the file
    name = tmp.name
    tmp_bed = pybedtools.BedTool(name)
    tmp_bed.sort().merge().saveas(tmp_merge.name)
    tmp_merge.close()

    E.info("indexing bed entries")
    # index the bed entries
    merged = IndexedGenome.Simple()
    for bed in Bed.iterator(IOTools.openFile(tmp_merge.name)):
        merged.add(bed.contig, bed.start, bed.end)

    counts = collections.defaultdict(int)
    # list of samples
    samples = options.infiles

    E.info("counting no. samples overlapping each interval")
    for sample in samples:
        found = set()
        for bed in Bed.iterator(IOTools.openFile(sample)):
            if merged.contains(bed.contig, bed.start, bed.end):
                key = [bed.contig] + \
                    [x for x in merged.get(bed.contig, bed.start, bed.end)]
                key = (key[0], key[1][0], key[1][1])
                if key in found:
                    continue
                found.add(key)

                # tuple of interval description as key - (contig, start, end)
                counts[key] += 1

    # open outfile
    options.stdout.write("contig\tstart\tend\tcount\n")

    E.info("outputting result")
    for interval, count in sorted(counts.items()):
        options.stdout.write(
            "\t".join(map(str, interval)) + "\t" + str(count) + "\n")

    # write footer and output benchmark information.
    E.Stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))
